#! /bin/sh
# \
exec tclsh${TCL} "$0" ${1+"$@"}
#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------
namespace eval ::thatch {
	if {[package provide thatch] ne {}} { return }

	proc thatchBootstrap {{fn ""}} {
		if {$fn eq ""} { set fn [info script] }
		if {[catch {
			set f [open $fn r]
			fconfigure $f -eofchar \x1a
			read $f
			seek $f 1 current
			set d [read $f]
			close $f
		} err]} {
			return 0
		}
		if {[string length $d] == 0} { return 0 }
		uplevel #0 $d
		package require thatch
	}
	package ifneeded thatch 0.5 [namespace current]::thatchBootstrap
}

package require thatch
::thatch::thatchSetup [info script] -nofile -noglob

source /bin/thatchtool.tcl
#-------------------------------------------------------------------

# THATCH /thatch/thatch.tcl size 27626 mtime 1662916831
#
# Thatch - Tcl Attach
#
#
# Stuart Cassoff
#
# May 2009
# December 2009
# January 2010
# October 2010
# March 2012
# Summer 2012
# Spring 2018
# Version 0.1
#
# Spring 2018
# Version 0.2
#
# Summer 2018
# Version 0.3
#
# Summer 2022
# Version 0.4
#
# Summer 2022
# Version 0.5
#


[namespace eval ::thatch {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	thatch
		title	Thatch
		descr	"App Builder"
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig importfrom [namespace current]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2009,2018,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}; # pkgconfig


#-------------------------------------------------------------------
# Thatch preamble and bootstrap code
#-------------------------------------------------------------------


proc preamble {{wish {0}}} {
	return "#! /bin/sh\n# \\\nexec [expr {$wish ? "wi" : "tcl"}]sh\${TCL} \"\$0\" \${1+\"\$@\"}"
}


variable bootstrap [string range {
#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------
namespace eval ::thatch {
	if {[package provide thatch] ne {}} { return }

	proc thatchBootstrap {{fn ""}} {
		if {$fn eq ""} { set fn [info script] }
		if {[catch {
			set f [open $fn r]
			fconfigure $f -eofchar \x1a
			read $f
			seek $f 1 current
			set d [read $f]
			close $f
		} err]} {
			return 0
		}
		if {[string length $d] == 0} { return 0 }
		uplevel #0 $d
		package require thatch
	}
	package ifneeded thatch @VER@ [namespace current]::thatchBootstrap
}

package require thatch
::thatch::thatchSetup [info script] @THATCHOPTS@

source @MAIN@
#-------------------------------------------------------------------
} 1 end-1]


#-------------------------------------------------------------------
# Other
#-------------------------------------------------------------------

proc linterleave {l1 l2} {
	set l {}
	foreach e1 $l1 e2 $l2 { lappend l $e1 $e2 }
	return $l
}

proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

proc aliasconst {name val args} {
	if {[llength $args] % 2 != 0} { error "must be %2!" }
	foreach {n v} [linsert $args 0 $name $val] {
		interp alias {} [imn $n] {} return -level 0 $v
	}
}

proc unknownOption {opt opts} {
	set msg "unknown option \"$opt\""
	if {[llength $opts] == 0} { return $msg }
	append msg ", must be: "
	append msg [lindex $opts 0]
	if {[llength $opts] == 2} {
		append msg " or [lindex $opts 1]"
	} elseif {[llength $opts] > 2} {
		append msg ", " [join [lrange $opts 1 end-1] ", "] " or " [lindex $opts end]
	}
	return $msg
}

#-------------------------------------------------------------------
# File
#-------------------------------------------------------------------

proc openItIn {fn {mode r} args} {
	if {$fn in {- stdin}} {
		set f stdin
	} else {
		if {[string index $fn 0] eq "@"} {
			set f [string range $fn 1 end]
		} else {
			set f [open $fn $mode]
		}
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
	return $f
}

proc openItOut {fn {mode w} args} {
	if {$fn in {- stdout}} {
		set f stdout
	} else {
		set f [open $fn $mode]
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
	return $f
}

proc closeIt {f} {
	if {$f in {stdin stdout}} {
		return {}
	}
	return [chan close $f]
}

proc inhaleIt {fn args} {
	return [read [set f [openItIn $fn r {*}$args]]][closeIt $f]
}

interp alias {} [imn openIt] {} [imn openItIn]

#-------------------------------------------------------------------
# Transforms
#-------------------------------------------------------------------

proc thatchWhatTransforms {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list copy base64 zip] $pat]]
}

if {[namespace ensemble exists binary] && "encode" in [dict keys [namespace ensemble configure binary -map]] && "decode" in [dict keys [namespace ensemble configure binary -map]]} {
	namespace eval ::base64 {
		proc encode {args} { binary encode base64 {*}$args }
		proc decode {args} { binary decode base64 {*}$args }
		package provide base64 2000
	}
}; # if binary encode/decode

namespace eval ::zip {
	proc encode {args} { zlib compress   {*}$args }
	proc decode {args} { zlib decompress {*}$args }
	package provide zip 2000
}

proc thatchTransform {data process args} {
	foreach alg $args {
		if {$alg eq "copy"} { continue }
		package require $alg
		set data [::${alg}::${process} $data]
	}
	return $data
}

#-------------------------------------------------------------------
# Checksums
#-------------------------------------------------------------------

proc thatchWhatChecksums {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list md5 sha256] $pat]]
}

proc thatchChecksum {data alg} {
	package require $alg
	set ns [string map [list md5 md5 sha256 sha2] $alg]
	return [::${ns}::${alg} -hex $data]
}

proc thatchMakeChecksums {data args} {
	set cksums {}
	foreach alg [lsort -nocase -dictionary $args] {
		set alg [string tolower $alg]
		lappend cksums $alg:[thatchChecksum $data $alg]
	}
	return [join $cksums ,]
}

proc thatchCheckChecksums {data args} {
	set errs {}
	foreach cksum [lsort -nocase -dictionary [split $args ,]] {
		lassign [split $cksum :] alg sum
		set alg [string tolower $alg]
		set csum [thatchChecksum $data $alg]
		if {$csum ne $sum} {
			lappend errs [list $cksum $csum]
		}
	}
	return $errs
}

#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#
#	i == file info
#	N == \n
#	Z == ^Z
#
#	ZN# THATCH iiiiiiiiiiiiiiiN
#	||__sig___||__file info__||
#	|          |file info line|
#	|_____file info chunk_____|
#
#
#	file info       fis
#	file info line  fil
#	file info chunk fic
#	file info dict  fid
#	file spec       fs
#
#-------------------------------------------------------------------


#
# File Spec (fs) fields
#
proc fsFields {} {
	return [list name size mtime base64 zip md5 sha256 remove rehome modbase]
}

proc fid->fis {fid} {
	set name [dict get $fid name]
	if {[dict exists $fid rehome]} {
		if {[dict get $fid rehome] ne ""} {
			set name [dict get $fid rehome]
		}
		dict unset fid name
		dict unset fid rehome
	}
	if {[dict exists $fid modbase] && [dict get $fid modbase] eq {}} {
		dict unset fid modbase
	}
	return "$name $fid"
}

proc fis->fid {fis} {
	set q [dict create name [lindex $fis 0] {*}[lrange $fis 1 end]]
}

proc fidFields {args} {
	return "name transforms checksums size stored-size mtime modbase"
}

proc newFid {args} {
	return [dict create ${*}[fidFields] {*}$args]
}

#
# File Spec (fs) -> File Info Dict (fid)
#
proc fs->fid {fs {dv {}}} {
	if {$dv ne ""} { upvar $dv data }
	lassign $fs {*}[fsFields]
	set transforms [list]
	set chanconf [dict create]

	# If zip & base64 do zip first
	if {$zip == 1} {
		lappend transforms zip
		dict set chanconf -translation binary
	}
	if {$base64 == 1} {
		lappend transforms base64
		dict set chanconf -translation binary
	}

	set fid [dict create name $name]

	if {[llength $transforms] > 0} { dict set fid transforms $transforms }

	if {$dv eq ""} {
		return $fid
	}

	set filename $name
	file stat $filename s
	set data [inhaleIt $filename {*}$chanconf]

	set algs [list]
	if {$md5 == 1}  { lappend algs md5 }
	if {$sha256 == 1} { lappend algs sha256 }

	set checksums [thatchMakeChecksums $data {*}$algs]
	set data [thatchTransform $data encode {*}$transforms]
	set s(stored-size) [string bytelength $data]

	if {$checksums ne ""} {
		dict set fid checksums $checksums
	}

	if {$size == 1} {
		dict set fid size $s(size)
		if {$s(size) != $s(stored-size)} {
			dict set fid stored-size $s(stored-size)
		}
	}

	if {$mtime == 1} {
		dict set fid mtime $s(mtime)
	}

	dict set fid rehome $rehome
	dict set fid modbase $modbase

	return $fid
}

#
# Writes a Thatch header
#
proc fid->fic {of fid} {
	puts -nonewline $of [^Z][sig][fid->fis $fid]\n
}

#
# Writes a File Info Chunk (fic): the Thatch header and the file
# from a File Spec (fs)
#
proc fs->fic {of fs} {
	set fid [fs->fid $fs data]
	fid->fic $of $fid
	puts -nonewline $of $data
}

#
# Arguments (ags) -> File Spec List (fsl)
#
proc ags->fsl {ags} {
	set fsl [list]
	#         name      var     val default
	set al [list {*}{
		{ -         doArgs    0  1 }
		{ allsize   allSize   1 -1 }
		{ allmtime  allMtime  1 -1 }
		{ allbase64 allBase64 1 -1 }
		{ allzip    allZip    1 -1 }
		{ allmd5    allMd5    1 -1 }
		{ allsha256 allSha256 1 -1 }
		{ noallsize   allSize   0 -1 }
		{ noallmtime  allMtime  0 -1 }
		{ noallbase64 allBase64 0 -1 }
		{ noallzip    allZip    0 -1 }
		{ noallmd5    allMd5    0 -1 }
		{ noallsha256 allSha256 0 -1 }
		{ size      size      1 -1 }
		{ mtime     mtime     1 -1 }
		{ base64    base64    1 -1 }
		{ zip       zip       1 -1 }
		{ md5       md5       1 -1 }
		{ sha256    sha256    1 -1 }
		{ nosize    size      0 -1 }
		{ nomtime   mtime     0 -1 }
		{ nobase64  base64    0 -1 }
		{ nozip     zip       0 -1 }
		{ nomd5     md5       0 -1 }
		{ nosha256  sha256    0 -1 }
		{ allremove allRemove 1  0 }
		{ remove    remove    1 -1 }
		{ file      argIsFs   1  0 }
		{ home      argIsHm   1  0 }
		{ rehome    argIsRHm  1  0 }
		{ modbase   argIsMb   1  0 }
		{ rename    argIsRen  1  0 }
	}]
	foreach a $al { set [lindex $a 1] [lindex $a 3] }

	set    homedir ""
	set  rehomedir ""
	set modbasedir ""
	set resetOpts 1
	set gopts {}
	# Arg options override FileSpec options
	foreach arg $ags {
		if {$resetOpts} {
			set resetOpts 0
			set rename ""
			foreach a $al {
				# Only reset specific options
				if {[lindex $a 0] ni [fsFields]} { continue }
				set [lindex $a 1] [lindex $a 3]
			}
		}
		set home    ""
		set rehome  ""
		set modbase ""
		if {$argIsHm} {
			set argIsHm 0
			set homedir $arg
			continue
		}
		if {$argIsRHm} {
			set argIsRHm 0
			set rehomedir $arg
			continue
		}
		if {$argIsMb} {
			set argIsMb 0
			set modbasedir $arg
			continue
		}
		if {$argIsRen} {
			set argIsRen 0
			set rename $arg
			continue
		}
		# Does arg look like an option or a FileSpec?
		if {$doArgs && !$argIsFs && [string index $arg 0] eq "-"} {
			# Valid options are assigned to vars and added to gopts
			if {[set i [lsearch -index 0 $al [string range $arg 1 end]]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
				if {[string range $arg 1 3] eq "all" || [string range $arg 1 5] eq "noall"} {
					lappend gopts [string range $arg 1 end]
				}
				continue
			}
		}
		set argIsFs 0
		# Check FileSpec for options
		set opts [lassign [split $arg %] name]
		foreach opt $opts {
			if {$argIsHm} {
				set argIsHm 0
				set home $opt
				continue
			}
			if {($opt ni $gopts || $opt eq "home") && [set i [lsearch -index 0 $al $opt]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
			}
		}

		# Empty name can be used to set global options
		if {$name eq ""} {
			set resetOpts 1
			continue
		}

		# Use defaults if options not supplied
		#foreach n {size mtime base64 zip md5 sha256 remove rehome modbase} {}
		foreach n {size mtime base64 zip md5 sha256 remove} {
			if {[set $n] == -1} { set $n [set all[string totitle $n]] }
		}

		if {$home eq ""} {
			set home [expr {[string index $name 0] eq "/" ? "" : $homedir}]
		}

		set file [file join $home $name]

		if {$rename ne ""} {
			set name $rename
		}
		if {$rehomedir ne ""} {
			set rehome [file join [expr {[string index $name 0] eq "/" ? "" : $rehomedir}] $name]
		}
		if {$modbasedir ne "" && [regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $file] -> pn pv]} {
			set modbase $modbasedir
		}

		lappend fsl [list $file $size $mtime $base64 $zip $md5 $sha256 $remove $rehome $modbase]

		set resetOpts 1
	}

	return $fsl
}

proc thatchOpen {fn {dir {in}}} {
	if {$dir eq "out"} {
		set mode w
		set eofc ""
	} else {
		set mode r
		set eofc [^Z]
	}
	return [openIt[string totitle $dir] $fn $mode -eofchar $eofc -encoding utf-8 -translation lf]
}

#
# Create or modify a Thatch file
#
proc thatchOp {op ifn ofn args} {
	set if [openItIn $ifn]
	if {$op eq "create"} {
		chan configure $if -eofchar [^Z]
	} else {
		chan configure $if -eofchar ""
	}
	set of [thatchOpen $ofn out]
	chan copy $if $of
	closeIt $if
	foreach fs [ags->fsl $args] {
		fs->fic $of $fs
	}
	closeIt $of
	return 1
}

proc thatchCopyCallbackMain {t fn f fid args} {
	lassign $args fsl of
	puts -nonewline $of [read $f]
	return skip
}

proc thatchCopyCallbackFile {t fn f fid args} {
	lassign $args fsl of

	set dosize 0
	set dobase64 0
	set undobase64 0
	set dozip 0
	set undozip 0
	set ofid $fid
	if {[set i [lsearch -index 0 $fsl [dict get $fid name]]] != -1} {
		if {[lindex $fsl $i [lsearch [fsFields] remove]]} {
			return ""
		}
		foreach fld [list size mtime] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0 && [dict exists $fid $fld]} {
				set fid [dict remove $fid $fld]
			} elseif {$v == 1} {
				if {$fld eq "size"} {
					set dosize 1
				}
			}
		}
		foreach fld [list base64 zip] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {![dict exists $fid transforms]} { continue }
				set transforms [dict get $fid transforms]
				if {[set i [lsearch $transforms $fld]] != -1} {
					set transforms [lreplace $transforms $i $i]
				}
				if {$fld eq "base64"} {set undobase64 1}
				if {$fld eq "zip"}    {set undozip 1}
			} elseif {$v == 1} {
				if {[dict exists $fid transforms]} { continue }
				lappend transforms $fld
				if {$fld eq "base64"} {set dobase64 1}
				if {$fld eq "zip"}    {set dozip 1}
			}
			if {[llength $transforms] == 0} {
				set fid [dict remove $fid transforms]
			} else {
				dict set fid transforms $transforms
			}
		}
		set checksums [dgwd {} $fid checksums]
		set sums1 [split $checksums :,]
		foreach fld [list md5 sha256] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {[set i [lsearch $sums1 $fld]] != -1} {
					set sums1 [lreplace $sums1 $i $i+1]
				}
			} elseif {$v == 1} {
			}
		}
		if {[llength $sums1] == 0} {
			if {[dict exists $fid checksums]} {
				set fid [dict remove $fid checksums]
			}
		} else {
			#dict set fid checksums something
		}
	}

	set d [read $f]

	if {$undobase64} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$undozip} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$dosize} {
		dict set fid size [string length $d]
	}
	if {$dobase64} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	if {$dozip} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	fid->fic $of [dict remove $fid offset]
	puts -nonewline $of $d

	return skip
}

proc thatchCopy {ifn ofn args} {
	set of [thatchOpen $ofn out]
	thatchTraverse $ifn [list [imn thatchCopyCallbackFile] [imn thatchCopyCallbackMain]] [ags->fsl $args] $of
	closeIt $of
	return 1
}

proc thatchNameMatch {fid g} {
	return [string match $g [dict get $fid name]]
}

proc zthatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		dict set res main [read $f]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res thIndex [dict get $fid name] $fid
		}
	}
	return ""
}

proc zthatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		set size [string length [read $f]]
		dict set res $fn [dict set fid size $size]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFileGlobCallback {t fn f fid args} {
	if {$t == 2} { return "" }
	variable res
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFileGlobCallback]] {*}$args
	return $res
}

proc thatchTraverse {fn cb args} {
	# This is supposed to be a Thatch file so it should conform
	set f [thatchOpen $fn]

	# Setup callbacks
	lassign $cb cb mcb
	if {$mcb eq ""} {
		set mcb $cb
	}

	# Get the 'main'
	set fid [dict create name $fn offset 0]
	set r [$mcb 2 $fn $f $fid {*}$args]
	if {$r ne "skip"} {
		read $f
	}

	# Note its largesse
	set offset [tell $f]

	# Now let's see if there's anything else
	while {1} {
		# We have hit eof so check if it's a ^Z eof or a real eof.
		# Do this by turning off -eofchar detection and reading one char:
		# it should be a ^Z. Set -eofchar back to ^Z after.
		fconfigure $f -eofchar {}
		set cz [read $f 1]
		# Real eof? If so, $cz should also be "".
		if {[eof $f] || $cz eq ""} {
			# Normal case
			if {[eof $f] && $cz eq ""} {
				break
			}
			# todo, strange case
			break
		}
		# Gotta be control-zee
		if {$cz ne [^Z]} { binary scan $cz c c; error "Expected ^Z at $offset, got \"[format %#x $c]\"::[gets $f]" }
		fconfigure $f -eofchar [^Z]
		# After the ^Z should be the sig
		set something [read $f [string length [sig]]]
		if {$something ne [sig]} {
			if {[eof $f]} {
				if {$something eq ""} {
					# null block
					continue
				}
			} else {
				append something [read $f]
			}
			# anon block
			continue
		}
		# Now we should be at the start of the file info line
		set fis [gets $f]
		# Do some checks ...
		if {[eof $f]} { explode }

		# Recalc offset:
		#    ^Z           \n                                 fis                               \n
		incr offset; incr offset [string length [sig]]; incr offset [string length $fis]; incr offset

		set fid [fis->fid $fis]
		dict set fid offset $offset
		set r [$cb 1 $fn $f $fid {*}$args]
		if {$r eq "stop"} { break }
		set size  [dgwd -1 $fid size]
		set ssize [dgwd -1 $fid stored-size]
		if {$ssize > 0} { set size $ssize }
		if {$size == -1} {
			if {$r ne "skip"} {
				read $f
			}
			set offset [tell $f]
		} else {
			if {$r ne "skip"} {
				seek $f $size current
			}
			incr offset $size
		}
	}
	closeIt $f
}

proc thatchExtract {fn fid} {
	set f [thatchOpen $fn]

	set name [lindex [dict keys $fid] 0]
	seek $f [dict get $fid $name offset] start

	set size [dgwd -1 $fid $name size]
	set ssize [dgwd -1 $fid $name stored-size]

	set transforms [dgwd {} $fid $name transforms]
	set checksums [dgwd "" $fid $name checksums]

	if {$ssize < 1} {
		set ssize $size
	}

	if {$ssize > 0} {
		set data [read $f $ssize]
	} else {
		set data [read $f]
	}

	closeIt $f

	set data [thatchTransform $data decode {*}[lreverse $transforms]]

	set checksum_errs [thatchCheckChecksums $data {*}$checksums]
	if {[llength $checksum_errs] > 0} {
		error "checksum mismatches ($fid):\n![join $checksum_errs \n]!"
	}

	return $data
}

proc thatchExtractMain {fn} {
	set f [thatchOpen $fn]
	set data [read $f]
	closeIt $f
	return $data
}

proc thatchIndexGlob {fn args} {
	variable thIndex
	if {[llength $args] == 0} { set args [list *] }
	set res [dict create]
	foreach a $args {
		set res [dict merge $res [dict filter $thIndex key $a]]
	}
	return $res
}

proc thatchCreateIndex {fn} {
	thatchDestroyIndex
	variable thIndex [thatchFileGlob $fn]
	interp alias {} [imn thatchGlob] {} [imn thatchIndexGlob]
}

proc thatchDestroyIndex {} {
	variable thIndex
	unset -nocomplain thIndex
	interp alias {} [imn thatchGlob] {} [imn thatchFileGlob]
}

#-------------------------------------------------------------------
# Overrides
#-------------------------------------------------------------------

proc thatchSourceOverride {fn args} {
	if {[dict size [set fid [thatchGlob $fn {*}$args]]] > 0} {
		set  script [info script]
		info script [lindex [dict keys $fid]]
		set  retval [uplevel 1 [thatchExtract $fn $fid]]]
		info script $script
	} else {
		set retval [uplevel 1 [list ::tcl::tcl_source {*}$args]]
	}
	return $retval
}

proc thatchImageOverride {fn args} {
	if {
		   [llength $args] > 3
		&& [lindex $args 0] eq "create"
		&& [set ind [lsearch $args -file]] != -1
		&& [dict size [set fid [thatchGlob $fn [lindex $args $ind+1]]]] > 0
	} {
		lset args $ind -data
		lset args $ind+1 [thatchExtract $fn $fid]
	}
	return [uplevel 1 [list ::tcl::tcl_image {*}$args]]
}

proc thatchLoadOverride {fn args} {
	set lfi -1
	foreach lfn $args {
		incr lfi
		if {[string index $lfn 0] ne "-"} {
			break
		}
	}
	if {$lfi > -1 && [dict size [set fid [thatchGlob $fn $lfn]]] > 0} {
		set c [file tempfile t thatch]
		chan configure $c -translation binary
		puts -nonewline $c [thatchExtract $fn $fid]
		close $c
		lset args $lfi $t
		if {[llength $args] > ($lfi - 1)} {
			lappend args [string totitle [lindex [regexp -inline {^lib([a-zA-z_]+).*} [file tail $lfn]] end]]
		}
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
		file delete -force $t
	} else {
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
	}
	return -code $code $res
}

proc thatchFileOverride {fn args} {
	set ok 0
	if {[llength $args] == 2} {
		set fcmd [lindex $args 0]
		if {$fcmd in {exists isfile owned readable executable isdirectory writable atime mtime attributes size type}} {
			set ok 1
		}
	}
	if {[llength $args] == 3} {
		set fcmd [lindex $args 0]
		if {$fcmd in {lstat stat}} {
			set ok 1
		}
	}

	if {[lindex $args 0] eq "type" && [lindex $args 1] eq ""} { return "directory" }
	if {$ok && [dict size [set fid [thatchGlob $fn [lindex $args 1]]]] == 1} {
		switch -exact -- [lindex $args 0] {
		exists -
		isfile -
		owned -
		readable {
			return 1
		}
		executable -
		isdirectory -
		writable {
			return 0
		}
		type {
			return "file"
		}
		atime -
		mtime {
			return [dict get $fid mtime]
		}
		attributes {
			return 0o444
		}
		size {
			return $size
		}
		lstat -
		stat {
			upvar 1 [lindex $args 2] a
			unset -nocomplain a
			array set a [list atime [dict get $fid mtime] ctime [dict get $fid mtime] dev -1 gid -1 ino -1 mode 0o444 mtime [dict get $fid mtime] nlink 1 size [dict get $fid size] type "file" uid -1]
			return ""
		}
		default { explode! }
		}
	}
	return [uplevel 1 [list ::tcl::tcl_file {*}$args]]
}

proc thatchGlobOverride {fn args} {
#puts $fn:$args
#if {[catch {info level -1} e]} { set e :: }
#puts $e:$args
	variable thIndex
	array set g {-join 0 -tails 0 -types {} -directory "" -path "" patterns {}}
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -- $arg {
			-join - -tails { set g($arg) 1 }
			-dir - -directory - -path - -types {
				if {$arg eq "-dir"} { set arg "-directory" }
				set state setelem
				set elem $arg
			}
			-nocomplain {}
			-- { set state endopts }
			default { lappend g(patterns) $arg }
			}
		}
		endopts { lappend g(patterns) $arg }
		setelem { set state opt; set g($elem) $arg }
		}
	}
	if {$g(-join)} { set g(patterns) [file join {*}$g(patterns)] }
	set fid [dict create]
	foreach p $g(patterns) {
		set fid [dict merge $fid [thatchGlob $fn [file join $g(-directory) $g(-path) $p]]]
	}
	if {[dict size $fid] > 0} {
		return [dict keys $fid]
	}
	tailcall ::tcl::tcl_glob {*}$args
}

#-------------------------------------------------------------------
# Setup
#-------------------------------------------------------------------

proc overrideAProc {fn proc} {
	if {[llength [info commands ::$proc]]} {
		rename ::$proc ::tcl::tcl_$proc
		interp alias {} ::$proc {} [imn thatch[string totitle $proc]Override] $fn
	} else {
		
	}
}

proc thatchSetup {fn args} {
	set oo {}; foreach o [set op {file glob image load source index packages modules}] { lappend oo -no$o }
	foreach a $args {
		if {[lsearch -exact $oo $a] == -1} {
			error [unknownOption $a $oo]
		}
		set i [lsearch -exact $op [string range $a 3 end]]
		set op [lreplace $op $i $i]
	}
	set createIndex 0
	if {[set i [lsearch -exact $op index]] != -1} {
		set op [lreplace $op $i $i]
		set createIndex 1
	}
	set setupPackages 0
	if {[set i [lsearch -exact $op packages]] != -1} {
		set op [lreplace $op $i $i]
		set setupPackages 1
	}
	set setupModules 0
	if {[set i [lsearch -exact $op modules]] != -1} {
		set op [lreplace $op $i $i]
		set setupModules 1
	}
	foreach p $op { overrideAProc $fn $p }
	if {$createIndex} { thatchCreateIndex $fn }
	if {$setupPackages} {
		dict for {k v} [thatchGlob $fn *pkgIndex.tcl *tclIndex] {
			set ::dir [file dirname $k]
			uplevel #0 [thatchExtract $fn [dict create $k $v]]
		}
		unset -nocomplain ::dir
	}
	if {$setupModules} {
		dict for {k v} [thatchGlob $fn {*-[0-9]*.tm}] {
			if {![regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $k] -> pn pv]} { continue }
			if {[dict exists $v modbase]} {
				set pn [string map {/ ::} [string trimleft [string range $k [string length [dict get $v modbase]] end-[string length [file tail $k]]] /]]$pn
			}
			package ifneeded $pn $pv "[::list package provide $pn $pv];[::list source -encoding utf-8 $k]"
		}
	}
}

#-------------------------------------------------------------------
# Package Setup
#-------------------------------------------------------------------

proc setup {} {
	package require Tcl 8.5
	aliasconst ^Z \x1a sig "\n# THATCH "
	interp alias {} [imn thatchCreate] {} [imn thatchOp] create
	interp alias {} [imn thatchAdd]    {} [imn thatchOp] add
	thatchDestroyIndex
	namespace export thatch*
	package provide thatch 0.5
}

namespace current}]::setup

# EOF

# THATCH /bin/thatchtool.tcl size 37685 mtime 1662921081
#! /bin/sh
# \
exec tclsh${TCL} "$0" ${1+"$@"}

#
# ThatchTool
#
# (formerly TkThatch)
#
#
# Stuart Cassoff
#
# December 2009, April 2010, Winter 2012, Summer 2012, Spring 2015, Winter 2018, Spring 2018
# Version 0.1
#
# Spring 2018
# Version 0.2
#
# Summer 2018
# Version 0.3
#
# Summer 2022
# Version 0.4
#
# Summer 2022
# Version 0.5
#

[namespace eval thatchtool {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	thatchtool
		title	ThatchTool
		descr	{Thatch Tool}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2009,2018 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


#---------------------------------------------------------------------
# Info
#---------------------------------------------------------------------

proc readme {} { return [string range {
<Control-q> will close windows or exit.

Home is optional.
Infile is optional.

To add files:
  Right-click in file name box and select a file.
  Erase mtime if mtime is not to be stored.
  Use checkbuttons to select whether to store size
  or enable base64 or zip transforms
  or md5 and sha256 chacksums,
  then click on the Keep button.

Thatchtool will load thatchtool project files (.tp)
from the command line.

"CSize" is Chunk Size.
Chunk Size is size of stored file + Thatch header inside the thatched file.

} 1 end-1]}

proc usage {} {
	return [format {%s%susage: %s [-?] [options] [filename] [...]} [cfg fullname] \n [string tolower [cfg name]]]
}

proc help {} {
	return [subst -nocommands [join {
		{-l  List (default)}
		{-c  Create}
		{-a  Add}
		{-r  Remove}
		{-e  Get}
		{-y  Copy}
		{-h  Home}
		{-H  Home + Thatch file home}
		{-o  fn Output file}
		{-m  fn Main}
		{-wp Wish preamble}
		{-np No preamble}
		{-pp Print preamble}
		{-pb Print bootstrap}
		{-tb Use Thatch bootstrap}
		{-v  Increase verbosity}
		{-?  This info}
	} \n]]
}

proc messages {} {
	return [join [cfg msgs] \n]
}

proc resetMessages {} {
	cfg msgs {}
}

proc log {msg {status 1}} {
	if {[catch {
		cfg+ msgs $msg
		if {[cfg guiCreated]} {
			if {$status} {
				setStatus $msg
			}
		} else {
			puts [cfg stdout] $msg
		}
	} err]} {
		#puts stderr $msg
	}
}

proc setStatus {msg} {}


#---------------------------------------------------------------------
# Misc
#---------------------------------------------------------------------

proc filterNv {data flds} {
	set res {}
	foreach l $data { if {[lindex $l 0] in $flds} { lappend res {*}$l } }
	return $res
}

proc mapCalc {ratio offset size width} {
	set res {}
	set sp [expr {int($offset * $ratio)}]
	set sx [expr {$sp % $width}]
	set sy [expr {$sp / $width}]
	set ep [expr {int(($offset + $size) * $ratio)}]
	set ex [expr {$ep % $width}]
	set ey [expr {$ep / $width}]

	if {$ey == $sy} {
		lappend res [list $sx $sy $ex $ey] {} {}
	} else {
		if {$sx == 0} {
			lappend res {}
		} else {
			lappend res [list $sx $sy [expr {$width + 1}] $sy]
			incr sy
		}
		if {$ey > $sy} {
			lappend res [list 0 $sy [expr {$width + 1}] $ey]
		} else {
			lappend res {}
		}
		if {$ex == $width} {
			lappend res {}
		} else {
			lappend res [list 0 $ey [expr {$ex + 1}] $ey]
		}
	}

	return $res
}

proc hexDump {d {w 8}} {
	set l [string length $d]
	set z {}
	for {set i 0; set ii [expr {$w - 1}]; set addr 0} {$i < $l} {incr i $w; incr ii $w; incr addr $w} {
		binary scan [string range $d $i $ii] @0H*@0a* hex ascii
		regsub -all -- {[^[:graph:] ]} $ascii {.} ascii
		regsub -all -- {..} $hex {& } hex
		lappend z [format "%04X  %-[expr {3 * $w}]s %-${w}s" $addr $hex $ascii]
	}
	return $z
}

proc moveFile {src dst} {
	if {1} {
	file rename -force $src $dst
	} else {
	file copy -force $src $dst
	file delete -force $src
	}
}


#---------------------------------------------------------------------
# Misc GUI
#---------------------------------------------------------------------

proc adjMainGeom {} {
	pjs!
	if {[dict size $pjs] == 0} {
		wm geometry . 300x100
	} elseif {[dict size $pjs] == 1} {
		wm geometry . {}
#		update idletasks
	}
	after idle [list [cfg NS]::centreToplevel .]
}

proc colTextWidth {tv txt} {
	return [expr {[font measure [ttk::style lookup [$tv cget -style] -font] $txt] + 30}]
}

proc adjTv {tv {cols {}}} {
	if {[llength $cols] == 0} {
		set cols [$tv columns]
	}
	foreach col $cols {
		set width [colTextWidth $tv [$tv heading $col -text]]
		foreach item [$tv children {}] {
			set txtwidth [colTextWidth $tv [$tv set $item $col]]
			if {$txtwidth > $width} {
				set width $txtwidth
			}
		}
		$tv column $col -width $width -minwidth $width
	}
}

proc autoscrollbar {w args} {
	set w [ttk::scrollbar $w {*}$args]
	::autoscroll::autoscroll $w
	return $w
}

proc changeWordWrap {} {
	set w [epmW]
	$w configure -wrap [expr {[$w cget -wrap] eq "none" ? "word" : "none"}]
}


#---------------------------------------------------------------------
# Dialog
#---------------------------------------------------------------------

proc oc {title message} {
	set choice [dialog [randWindow] $title $message "" 1 "&Ok" &Cancel"]
	if {$choice == -1} { set choice end }
	return [string index OC $choice]
}

proc yn {title message} {
	set choice [dialog [randWindow] $title $message "" 1 "&Yes" "&No"]
	if {$choice == -1} { set choice end }
	return [string index YN $choice]
}

proc okcancel {title message} {
	return [expr {[oc $title $message] eq "O"}]
}

proc yesno {title message} {
	return [expr {[yn $title $message] eq "Y"}]
}

proc dialog {window title text bitmap default args} {
	set ns [cfg NS]:
	destroyToplevel $window
	newToplevel $window
	groupToplevel $window [gui W]
	setToplevelTitle $window $title
	transientizeToplevel $window [gui W]
	set t [split $text \n]
	set weegs [dict create]
	eval [gub "
		[gubMultiGob [llength $t] L F]
		.{L} -text $t
		.{B} -command [lmap i [::gub::ExpandRange 0-[expr {[llength $args]-1}]] { list $ns:dialogClick $i}]
		;F [string repeat {B } [llength $args]]
		[gubButtonGob -buttonchar B -modifiers Alt {*}$args]
		[expr {$default > -1 ? "/B" : ""}]
	" $window weegs]
	foreach k {Control-q Control-Q Escape} { bind $window <$k> [list $ns:dialogClick -1] }
	onToplevelDelete $window [list $ns:dialogClick -1]
	centreToplevel $window [gui W]
#	wm attributes $window -topmost 1
	grab set $window
	if {$default > -1} { focus [dict get $weegs B$default] }
	gui dialogRetval -1
	vwait [gui& dialogRetval]
	grab release $window
	destroy $window
	return [gui dialogRetval]
}

proc dialogClick {what} {
	gui dialogRetval $what
}


#---------------------------------------------------------------------
# Messages
#---------------------------------------------------------------------

proc clearMessages {t} {
	$t delete 1.0 end
	resetMessages
}

proc viewMessages {} {
	set w .q
	destroy $w
	set w [newToplevel $w]
	groupToplevel $w [gui W]
	set weegs [dict create]
	eval [gub "
		;F t S
		;  C
		;F B B
		.B0 -command {[list after idle [list destroy $w]]}
		.B1 -command {[cfg NS]::clearMessages /t0/}
		:F0,t0
		-.,F0 ; |.,F0
		/t0 viewText
		/B0 okButton
		[gubButtonGob -buttonchar B &Ok &Clear]
		&B0 Control+q
		&B0 Escape
#		/. W
	" $w weegs]
	setToplevelTitle $w [string cat [getToplevelTitle [gui W]] " - Messages"]
	[dict get $weegs viewText] insert end [messages]
	focus [dict get $weegs okButton]
	centreToplevel $w [gui W]
}


#---------------------------------------------------------------------
# Config
#---------------------------------------------------------------------

proc configAdj {} {
	if {[cfg theme] eq ""} { cfg theme default }
	#set cfg(configFile) [file normalize $cfg(configFile)]
}

proc configLoad {{fn {?}}} {
	if {$fn eq {}} { set fn [cfg configFile] }
	if {![lassign [loadListFile $fn -filetypes [list [dict get [cfg filetypes] config]]] fn data]} { return 0 }
	dict for {k v} [filterNv $data [cfg configFileFields]] { cfg $k $v }
	cfg configFile $fn
	configAdj
	return 1
}

proc configSave {{fn {}}} {
	if {$fn eq {}} { set fn [cfg configFile] }
	foreach n [cfg configFileFields] { lappend l [list $n [cfg $n]] }
	if {![lassign [saveFile $fn [join $l \n]\n -filetypes [list [dict get [cfg filetypes] config]]] fn]} { return 0 }
	cfg configFile $fn
	configAdj
	return 1
}

proc autoloadConfig {} {
	configAdj
	#if {[info exists cfg(pConfigFile)]} { set cfg(configFile) $cfg(pConfigFile) }
	configAdj
	if {[cfg autoloadConfigFile]} { configLoad {} }
	configAdj
}

proc cfgThemeChanged {w} {
	ttk::style theme use [$w get]
}

proc configGo {} {
	::aitcs::aitcs new ecfg
	foreach f [cfg configBoolFields] { ecfg $f [cfg $f] }
	set wugs [createConfigGui]
	cfg->gui $wugs
	grab set [dict get $wugs W]
	focus [dict get $wugs B6]
}

proc configStop {wugs} {
	ttk::style theme use [cfg theme]
	grab release [dict get $wugs W]
	after idle [list destroy [dict get $wugs W]]
	ecfg---
}

proc configSaveStop {wugs} {
	foreach f [cfg configBoolFields] { cfg $f [ecfg $f] }
	gui->cfg $wugs
	configSave
	configStop $wugs
}

proc cfg->gui {wugs} {
	foreach p [cfg autoloadableProjects] {
		[dict get $wugs autoloadableProjects] insert {} end -values $p
	}
	[dict get $wugs themeSelector] current [lsearch [ttk::style theme names] [cfg theme]]
}

proc gui->cfg {wugs} {
	cfg autoloadableProjects {}
	foreach item [[dict get $wugs autoloadableProjects] children {}] {
		cfg+ autoloadableProjects [dict values [[dict get $wugs autoloadableProjects] set $item]]
	}
	cfg theme [[dict get $wugs themeSelector] get]
}

proc configAddProject {wugs} {
	set fn [tk_getOpenFile -filetypes [list [dict get [cfg filetypes] tp]] -parent [dict get $wugs W]]
	if {$fn eq ""} { return }
	[dict get $wugs autoloadableProjects] insert {} end -values [list $fn on]
}

proc configRemoveProject {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	[dict get $wugs autoloadableProjects] delete $item
}

proc configMoveProjectUp {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	set index [[dict get $wugs autoloadableProjects] index $item]
	if {$index == 0} { return }
	[dict get $wugs autoloadableProjects] move $item {} [incr index -1]
}

proc configMoveProjectDown {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	set index [[dict get $wugs autoloadableProjects] index $item]
	if {$index == [llength [[dict get $wugs autoloadableProjects] children {}]]} { return }
	[dict get $wugs autoloadableProjects] move $item {} [incr index]
}

proc configToggleProjectAutoload {wugs} {
	set item [[dict get $wugs autoloadableProjects] selection]
	if {$item eq ""} { return }
	[dict get $wugs autoloadableProjects] set $item autoload [expr {[[dict get $wugs autoloadableProjects] set $item autoload] ? "off" : "on"}]
}

proc createConfigGui {} {
	set w [newToplevel]
	groupToplevel $w [gui W]
	setToplevelTitle $w Configure
	set alpcols {project autoload}
	set weegs [dict create]
	eval [gub "
		;M Vp
		;  FpB B B B B
		;M LaKaLtXt
		;M KsKmKbKzKdKh
		;F B B
		.{M} -text {Autoloadable Projects} Options {Add Files Defaults}
		.Vp -columns {$alpcols} -show headings
		.{La,Lt} -text {Autoload Projects} Theme
		.Xt -state readonly -values {[ttk::style theme names]}
		.{K<smbzdh>} -text Size Mtime Base64 Zip Md5 Sha256
		.{K<asmbzdh>} -variable [lmap f [cfg configBoolFields] { ecfg& $f }]
		:M,V
		:L,K -sticky w
		:X -sticky ew -padx {4 0}
		-.,M0 ; |.,M0
		|M1 1
		[gubButtonGob -buttonchar B -modifiers Alt \
			&Add &Remove {Move &Up} {Move &Down} {&Toggle Autoload} &OK &Cancel \
		]
		/Vp autoloadableProjects
		/Xt themeSelector
		/B
		/. W
	" $w weegs]
	set tv [dict get $weegs autoloadableProjects]
	foreach col $alpcols {
		$tv heading $col -text [string totitle $col]
	}
	$tv column autoload -anchor center -stretch 0
	adjTv $tv $alpcols
	foreach b {0 1 2 3 4 5 6} cmd {configAddProject configRemoveProject configMoveProjectUp configMoveProjectDown configToggleProjectAutoload configSaveStop configStop} {
		[dict get $weegs B$b] configure -command [list [cfg NS]::$cmd $weegs]
	}
	bind [dict get $weegs themeSelector] <<ComboboxSelected>> [list [cfg NS]::cfgThemeChanged [dict get $weegs themeSelector]]
	foreach k {Control-q Control-Q Escape} { bind $w <$k> [list [cfg NS]::configStop $weegs] }
	onToplevelDelete $w [list [cfg NS]::configStop $weegs]
	centreToplevel $w [gui W]
	return $weegs
}


#---------------------------------------------------------------------
# Gui
#---------------------------------------------------------------------

proc createGui {{W {.}}} {
	pjs!;gui!
	set ns [cfg NS]:
	gui w [string trimright [gui W $W] .]
	gubWidgetMap [dict merge [gubWidgetMap] [dict create C [dict merge [dict get [gubWidgetMap] S] [dict create command $ns:autoscrollbar]]]]
	gui M [createMainMenu [gui w]]
	eval [gub -a "N ; : ; -. ; |. ; /N0 mainBook; &. Control+q $ns:fileExit" [gui w]]
	onToplevelDelete [gui W] $ns:fileExit
	catch {ttk::style theme use [cfg theme]}
	#ttk::style configure Treeview -rowheight [expr {[font metrics [ttk::style lookup Treeview -font] -linespace] + 2}]
	wm title [gui W] [cfg name]\ [cfg ver]
	[gui W] configure -menu [gui M]
	gui CVM [createChangeViewMenu [gui w]]
	epmCreateMenu [gui w]
	adjMainGeom
	#foreach i {3 4} { $gui(mainMenu,File) entryconfigure $i -state disabled }
}

proc createMainMenu {w} {
	set M $w.mMain
	set mob "
		.&File
		. &Configure	-command configGo
#		. &Messages	-command viewMessages
		. &Program Info	-command createProgramInfoGui
		. E&xit
		.&Project
		. &Open
		. &Save
		. Save &As	-command {[cfg NS]::projectSave ?}
		. &Close
		. &New
		.&Thatch
		. &Open		-command openThatch
	"
	eval [dict get [meb -menuname $M -namespace [cfg NS] $mob] tcl src]
	return $M
}

proc createChangeViewMenu {w} {
	set M $w.mChangeView
	set mob "
		.&Word Wrap -variable ::ww -command changeWordWrap
	"
	eval [dict get [meb -menuname $M -namespace [cfg NS] $mob] tcl src]
	return $M
}

proc createProgramInfoGui {} {
	set w .pinfo
	destroyToplevel $w
	set w [newToplevel $w]
	groupToplevel $w [gui W]
	setToplevelTitle $w "[cfg fullname] - Info"
	onToplevelDelete $w [list set ::ww $::ww]
	set weegs [dict create]
	eval [gub "
		;N FrtrC
		;    C
		;  FhthC
		;    C
		;  Fa
		;B
		[gubButtonGob -buttonchar B -modifiers Alt &Ok]
		.t -width 76 -height 20 -wrap none
		.B0 -command {after idle destroy /./}
		+{F<rha>} -text &Readme &Help &About
		:N0,t
		|.,F<rha> ; -.,F<rha>
#		&N0 <<NotebookTabChanged>> [cfg NS]::programInfoAdjustChangeViewMenuForText %W
		&B0 Control+q
		&B0 Escape
		/N0 book
		/B0 okbutton
		/Fa about
		/t<rh>
	" $w weegs]
	::ttk::notebook::enableTraversal [dict get $weegs book]
	setTextText [dict get $weegs tr] [readme]
	setTextText [dict get $weegs th] [usage]\n[help]
	::central::extras::CentralPackageBrowser new [dict get $weegs about]
	foreach t [list tr th] {
		[set t [dict get $weegs $t]] configure -state disabled 
		epmBindMenuTo $t [gui CVM]
	}
	focus [dict get $weegs okbutton]
	centreToplevel $w [gui W]
}

proc programInfoAdjustChangeViewMenuForText {w} {
	set ::ww [expr {[[$w select].t[string index [$w select] end] cget -wrap] eq "word"}]
}


#---------------------------------------------------------------------
# Project / Thatch
#---------------------------------------------------------------------

proc autoloadProjects {} {
	if {![cfg autoloadProjects]} { return }
	foreach p [cfg autoloadableProjects] {
		if {[lindex $p 1]} {
			projectOpen [lindex $p 0]
		}
	}
}

proc newProject {} {
	pjs!
	for {set i 0} {[dict exists $pjs [set id pj#$i]]} {incr i} {}
	dict set pjs $id [dict merge [cfg projectFields] [dict create id $id projectFile [string map {# ""} $id].tp et {} view {}]]
	return $id
}

proc addProjectToGUI {id} {
	pjs!
	adjMainGeom
	createProjectGUI [gui mainBook] $id
	[gui mainBook] add [set w [dict get $pjs $id w]] -text [dict get $pjs $id name]
	[gui mainBook] select $w
	return $id
}

proc projectOpen {{fn ""}} {
	pjs!
	if {![lassign [loadListFile $fn -filetypes [list [dict get [cfg filetypes] tp]]] fn data]} { return 0 }
	set id [newProject]
	dict set pjs $id [dict merge [dict get $pjs $id] [filterNv $data [dict keys [cfg projectFields]]]]
	dict set pjs $id projectFile $fn
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc projectSave {{fn {}}} {
	pjs!
	set id [getPjIdOfCurrentTab]
	gui->project $id
	if {$fn eq {}} {
		set fn [dict get $pjs $id projectFile]
	}
	if {$fn eq {}} {
		set fn ?
	}
	set l {}
	foreach f [cfg projectEditFields] {
		lappend l [list $f [dict get $pjs $id $f]]
	}
	if {![lassign [saveFile $fn [join $l \n]\n -filetypes [list [dict get [cfg filetypes] tp]]] fn]} { return 0 }
	dict set pjs $id projectFile $fn
	project->gui $id
	return 1
}

proc projectClose {} {
	pjs!
	set id [getPjIdOfCurrentTab]
	[gui mainBook] forget [dict get $pjs $id w]
	destroy [dict get $pjs $id w]
	dict unset pjs $id
	adjMainGeom
}

proc openThatch {{fn ""}} {
	pjs!
	if {$fn eq {} || $fn eq {?}} {
		gui!
		lappend filetypes [dict get [cfg filetypes] all]
		set fn [tk_getOpenFile -filetypes $filetypes -parent $gui(W) -title "Open Thatch"]
	}
	if {$fn eq {}} { return }
	set id [newProject]
	dict set pjs $id name $id
	dict set pjs $id outFile $fn
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc adjThatch {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {[file exists $fn]} {
		dict set pjs $id outFileSize [file size $fn]
		dict set pjs $id canvMapRatio [expr {(double([cfg mapCanvWidth] * [cfg mapCanvHeight])) / double([dict get $pjs $id outFileSize])}]
	} else {
		dict set pjs $id outFileSize 0
		dict set pjs $id canvMapRatio 0
	}
}

proc adjLineForDisp {id l} {
	pjs!
	dict set pjs $id mtimeFormat "%Y/%m/%d %T"
	if {[set f [dict get $pjs $id mtimeFormat]] ne ""} {
		set i [lsearch [cfg thatchTvFields] mtime]
		if {[set time [lindex $l $i]] ne ""} {
			lset l $i [clock format $time -format $f]
		}
	}
	return $l
}

proc project->gui {id} {
	pjs!
	foreach f [cfg projectEditFields] {
		[dict get $pjs $id w_$f] configure -validate none
		[dict get $pjs $id w_$f] delete 0 end
		[dict get $pjs $id w_$f] insert end [dict get $pjs $id $f]
		[dict get $pjs $id w_$f] configure -validate key
	}
	setProjectTitle $id [dict get $pjs $id name]
}

proc setProjectTitle {id t} {
	pjs!
	[gui mainBook] tab [dict get $pjs $id w] -text $t
}

proc gui->project {id} {
	pjs!
	foreach f [cfg projectEditFields] {
		dict set pjs $id $f [[dict get $pjs $id w_$f] get]
	}
	setProjectTitle $id [dict get $pjs $id name]
}

proc thatch->gui {id {focus {}}} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn] } {return }
	set offset [dict get $pjs $id outFileSize]
	set tv [dict get $pjs $id w_tv]
	set c [dict get $pjs $id w_c]
	set tf [lrange [cfg thatchTvFields] 1 end-1]
	set pos [file size $fn]
	set thIndex [::thatch::thatchFullFileGlob $fn]
	dict set pjs $id thIndex $thIndex
	$tv delete [$tv children {}]
	$c delete all
	[dict get $pjs $id w_ename] delete 0 end
	[dict get $pjs $id w_t] delete 1.0 end
	defaults->editThatchedFileInfoGui $id
	dict set pjs $id fishMap [dict create]
	foreach {v k} [lreverse [dict get $pjs $id thIndex]] {
		set l [list $k]
		if {[dict exists $v checksums]} {
			set x [dict get $v checksums]
			set r {}
			foreach {q z} [split $x ,:] {
				lappend r $q
			}
			dict set v checksums $r
		}
		foreach f [list size mtime transforms checksums offset] {
			lappend l [dgwd "" $v $f]
		}
		set csize [expr {$pos - [dict get $pjs $id thIndex $k offset]}]
		dict set pjs $id $fn $k csize $csize
		incr pos -$csize
		lappend l $csize
		dict set pjs $id fishMap [$tv insert {} 0 -values [adjLineForDisp $id $l]] $k
		createMapRegion $id $fn $k $c
	}
	$c itemconfigure all -fill [cfg mapCanvBackground]
	adjTv $tv [cfg thatchTvFields]
	if {$focus ne {}} {
		$tv focus $focus
	}
}

proc thatchedFile->gui {w id {view {}}} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn]} { return }
	dict set pjs $id view $view
	set sel [lindex [$w selection] 0]
	set name [dict get [dict get $pjs $id thIndex [dict get $pjs $id fishMap $sel]] name]
	set ext [file extension $name]
	set t [dict get $pjs $id w_t]
	set c [dict get $pjs $id w_c]
	thatchedFileInfo->editThatchedFileInfoGui $id
	set d [thatchExtract $fn [dict create $name [dict get $pjs $id thIndex $name]]]
	$t configure -state normal
	image delete {*}[$t image names]
	$t delete 1.0 end
	if {$ext in {".gif" ".png" ".ppm" ".xbm"}} {
		if {$view eq "hex"} {
			$t insert end [join [hexdump $d 32] \n]
			set title Img
		} else {
			$t image create end -image [image create [expr {$ext eq ".xbm" ? "bitmap" : "photo"}] -data $d]
			set title Hex
		}
	} else {
		set xfrms [dgwd {} $pjs $id thIndex $name transforms]
		if {"base64" in $xfrms} {
			if {$view eq {}} {
				set d [join [hexdump $d 32] \n]
			}
			set title [expr {$view eq {} ? "Txt" : "Hex"}]
		} else {
			if {$view eq "hex"} {
				set d [join [hexdump $d 32] \n]
			}
			set title [expr {$view eq {} ? "Hex" : "Txt"}]
		}
		$t insert end $d
	}
	$c itemconfigure all -fill [cfg mapCanvBackground]
	foreach item [$w selection] {
		set tag [dict get $pjs $id thIndex [dict get $pjs $id fishMap $item] name]
		$c itemconfigure $tag -fill [cfg mapCanvForeground]
		$c raise $tag
	}
	$t configure -state disabled
	[dict get $pjs $id w_bchgvw] configure -text [string cat \u2193 $title]
}

proc thatchedFileInfo->editThatchedFileInfoGui {id} {
	pjs!
	set w [dict get $pjs $id w_tv]
	set fi [$w set [lindex [$w selection] 0]]
	foreach tf [list name mtime] gf [list w_ename w_emtime] {
		[dict get $pjs $id $gf] delete 0 end
		[dict get $pjs $id $gf] insert end [dict get $fi $tf]
	}
	set x [dict get $fi transforms]
	foreach tf [list base64 zip] gf [list w_ebase64 w_ezip] {
		[dict get $pjs $id $gf] state [expr {$tf in $x ? "" : "!"}]selected
	}
	set x [dict get $fi checksums]
	foreach tf [list md5 sha256] gf [list w_emd5 w_esha256] {
		[dict get $pjs $id $gf] state [expr {$tf in $x ? "" : "!"}]selected
	}
	foreach tf [list size] gf [list w_esize] {
		set s [expr {[dict get $fi $tf] eq "" || [dict get $fi offset]  == 0 ? "!" : ""}]
		[dict get $pjs $id $gf] state ${s}selected
	}
}

proc defaults->editThatchedFileInfoGui {id} {
	pjs!
	foreach cf [list addDefSize addDefBase64 addDefZip addDefMd5 addDefSha256] gf [list w_esize w_ebase64 w_ezip w_emd5 w_esha256] {
		[dict get $pjs $id $gf] state [expr {[cfg $cf]?"":"!"}]selected
	}
	if {![cfg addDefMtime]} {
		[dict get $pjs $id w_emtime] delete 0 end
	}
}

proc getEditThatchedFileInfoGuiValues {id} {
	pjs!
	set d [dict create]
	foreach f [list name mtime] gf [list w_ename w_emtime] {
		dict set d $f [[dict get $pjs $id $gf] get]
	}
	foreach tf [list size base64 zip md5 sha256] gf [list w_esize w_ebase64 w_ezip w_emd5 w_esha256] {
		dict set d $tf [expr {"selected" in [[dict get $pjs $id $gf] state]}]
	}
	return $d
}

proc projectNew {} {
	pjs!
	set id [newProject]
	dict set pjs $id name $id
	adjThatch $id
	addProjectToGUI $id
	project->gui $id
	thatch->gui $id
	return $id
}

proc getPjIdOfCurrentTab {} {
	pjs!
	set w [[gui mainBook] select]
	dict for {id pj} $pjs {
		if {[dict get $pj w] eq $w} {
			return $id
		}
	}
	return {}
}

proc projectFieldChangedCallback {id} {
	pjs!
	dict set pjs $id et {}
	gui->project $id
}

proc projectFieldChanged {P} {
	pjs!
	set id [getPjIdOfCurrentTab]
	if {[dict get $pjs $id et] eq {}} {
		dict set pjs $id et [after idle [list [cfg NS]::projectFieldChangedCallback $id]]
	}
	return true
}

proc createProjectGUI {W id} {
	pjs!
	set ns [cfg NS]:
	set flds [cfg projectEditFields]
	set w [ttk::frame [randWindow $W.fF_]]
	set gob "
		;F c L E
		[string repeat ";    L E\n" [expr {[llength $flds]-1}]]
		;P F V C
		;    C
		;  F F L E L E K K K K K B B B Bv
		;    F t C
		;      C
		.c0 -width [cfg mapCanvWidth] -height [cfg mapCanvHeight] -background [cfg mapCanvBackground] -highlightthickness 0
		.{L} -text [lmap f $flds {string totitle $f}]
		.{L[expr {[llength $flds]}],[expr {[llength $flds]+1}]} -text Name Mtime
		.E0-[expr {[llength $flds]-1}] -validate key -validatecommand {$ns:projectFieldChanged %P}
		.{K} -text Size Base64 Zip Md5 Sha256
		.V0 -columns {[cfg thatchTvFields]} -show headings
		[gubButtonGob -buttonchar B -modifiers Alt Keep Remove Def \u2193Hex]
		.B0 -command {$ns:alterThatch $id}
		.B1 -command {$ns:removeFileFromThatch $id}
		.B2 -command {$ns:defaults->editThatchedFileInfoGui $id}
		.Bv -command {$ns:changeView $id} -width -5
		.B0-2 -width 0
		.t0 -height 10 -state disabled -wrap none
		:E -sticky ew
		:L -sticky w
		:c0 -rowspan [llength $flds]
		:F0,3,4,P0,V0,t0
		|.,F1,2,4;-F1,4
		|F0 2
		|F3 1
		-.,F0,2 1
		/c0 $id w_c
		/{E} {} [lmap f $flds { string cat $id " " w_ $f } ] {$id w_ename} {$id w_emtime}
		/V0 $id w_tv
		/t0 $id w_t
		/{K} {} {$id w_esize} {$id w_ebase64} {$id w_ezip} {$id w_emd5} {$id w_esha256}
		/Bv $id w_bchgvw
		/. {$id w}
	"
	eval [gub $gob $w pjs]

	set tv [dict get $pjs $id w_tv]
	foreach col [cfg thatchTvFields] {
		$tv heading $col -text [string totitle $col]
		$tv column $col -stretch 0 -anchor e
	}
	$tv column [lindex [cfg thatchTvFields] 0] -stretch 1 -anchor w
	$tv column [lindex [cfg thatchTvFields] end-1] -stretch 0 -anchor w
	adjTv $tv [cfg thatchTvFields]
	bind $tv <<TreeviewSelect>> [list $ns:thatchedFile->gui %W $id]
	foreach f [concat $flds [list ename emtime]] {
		epmBindMenuTo [dict get $pjs $id w_$f] [epmMenu]
	}
	foreach f [list esize ebase64 ezip emd5 esha256] {
		[dict get $pjs $id w_$f] state !alternate
	}
	epmBindMenuTo [dict get $pjs $id w_t] [gui CVM]

	return $id
}

proc createMapRegion {id fn k c} {
	pjs!
	lassign [lrepeat 3 [list]] topLine midRect botLine
	set csize [dict get $pjs $id $fn $k csize]
	incr csize -1
	while {[llength $topLine] < 1 && [llength $midRect] < 1 && [llength $botLine] < 1} {
		lassign [mapCalc [dict get $pjs $id canvMapRatio] [dict get $pjs $id thIndex $k offset] [incr csize] [cfg mapCanvWidth]] topLine midRect botLine
	}
	if {[llength $topLine] > 0} { $c create line $topLine -tag $k }
	if {[llength $botLine] > 0} { $c create line $botLine -tag $k }
	if {[llength $midRect] > 0} { $c create rectangle $midRect -outline {} -tag $k }
}

proc changeView {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {![file exists $fn]} { return }
	set w [dict get $pjs $id w_tv]
	set sel [lindex [$w selection] 0]
	if {[llength $sel] == 0} { return }
	set name [dict get [dict get $pjs $id thIndex [dict get $pjs $id fishMap $sel]] name]
	set view [dict get $pjs $id view]
	if {[file extension $name] in {".gif" ".png" ".ppm" ".xbm"}} {
		if {$view eq {}} {
			set view hex
		} else {
			set view {}
		}
		thatchedFile->gui $w $id $view
	} elseif {[file extension $name] in {".tcl" ".tm" ".txt" ".man" ".n" ".html" ".terms"}} {
		if {$view eq {}} {
			set view hex
		} else {
			set view {}
		}
		thatchedFile->gui $w $id $view
	}
}


#---------------------------------------------------------------------
# Thatch manip
#---------------------------------------------------------------------

proc alterThatch {id} {
	pjs!
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	if {$fn eq ""} { return }
	set gd [getEditThatchedFileInfoGuiValues $id]
	set name [dict get $gd name]
	if {$name eq ""} { return }
	set edit [dict exists $pjs $id thIndex $name]
	set ok 0
	set opts {}
	if {$edit} {
		set fid [dict get $pjs $id thIndex $name]
		foreach tf [list size] {
			set gv [dict get $gd $tf]
			set ke [dict exists $fid $tf]
			if {$ke == $gv} { continue }
			lappend opts -[expr {$gv ? "" : "no"}]$tf
		}
		foreach tf [list transforms checksums] stfl [list [list base64 zip] [list md5 sha256]] {
			set v [split [dgwd {} $fid $tf]]
			foreach f $stfl {
				set gv [dict get $gd $f]
				set ke [expr {$f in $v}]
				if {$ke != $gv} {
					lappend opts -[expr {$gv ? "" : "no"}]$f
				}
			}
		}
		file tempfile tmpfile thatch
		set ok [thatchCopy $fn $tmpfile {*}[join $opts] $name]
	} else {
		foreach tf [list size base64 zip md5 sha256] {
			set gv [dict get $gd $tf]
			lappend opts -[expr {$gv ? "" : "no"}]$tf
		}
		file tempfile tmpfile thatch
		set ok [thatchAdd $fn $tmpfile {*}[join $opts] $name]
	}
	if {$ok} {
		moveFile $tmpfile $fn
		adjThatch $id
		thatch->gui $id
	}
}

proc removeFileFromThatch {id} {
	pjs!
	set w [dict get $pjs $id w_tv]
	set sel [$w selection]
	if {[llength $sel] == 0} { return }
	set S [expr {[llength $sel] > 1 ? "s" : ""}]
	set msg {}
	set copyArgs {}
	foreach s $sel {
		set name [dict get [$w set $s] name]
		lappend msg $name
		lappend copyArgs -remove $name
	}
	if {![yesno "Remove File$S" [join [linsert $msg 0 "Remove File$S?"] \n]]} { return }
	set fn [file join [dict get $pjs $id home] [dict get $pjs $id outFile]]
	file tempfile tmpfile thatch
	if {![thatchCopy $fn $tmpfile {*}[join $copyArgs]]} { return }
	moveFile $tmpfile $fn
	adjThatch $id
	thatch->gui $id
}

proc preamble {} {
	return [::thatch::preamble [cfg wishpreamble]]
}

proc boot {} {
	return [string map [list @VER@ [package provide thatch] @THATCHOPTS@ [cfg thatchbootopts] @MAIN@ [cfg main]] $::thatch::bootstrap]
}

proc cgogo {} {
	switch -exact [cfg action] {
	list {
		set d [thatchGlob [cfg inFile]]
		if {[cfg verbosity] > 1} {
			dict for {k v} $d {
				puts stdout "$k $v"
			}
		} else {
			dict for {k v} $d {
				puts stdout $k
			}
		}
	}
	create {
		if {[cfg thatchboot]} {
			cfg cliFiles [list [cfg inFile] {*}[cfg cliFiles]]
			foreach c [chan pipe] n [list r w] {
				chan configure $c -encoding utf-8 -translation lf
				set $n $c
			}
			cfg inFile @$r
			if {[cfg preamble]} {
				puts $w [preamble]
			}
			puts $w [boot]
			close $w
		}
		thatch[string totitle [cfg action]] [cfg inFile] [cfg outFile] {*}[cfg cliFiles]
	}
	add - copy {
		thatch[string totitle [cfg action]] [cfg inFile] [cfg outFile] {*}[cfg cliFiles]
	}
	get {
#puts " [cfg inFile] [cfg cliFiles]"
		set res {}
		dict for {k v} [thatchGlob [cfg inFile] [cfg cliFiles]] {
			lappend res [thatchExtract [cfg inFile] [dict create $k $v]]
		}
		puts stdout [join $res \n]
	}
	}

	foreach fn [cfg toDel] {
		file delete -force $fn
	}
}


#---------------------------------------------------------------------
# Setup / Startup / Quit
#---------------------------------------------------------------------

proc fileExit {} {
	#exit
	destroy .
}

proc resolveThatch {fileNames} {
	set fns [list]
	foreach fn $fileNames {
		if {$fn ne ":thatch:"} {
			lappend fns $fn
			continue
		}
		if {[dict size [set fid [thatchGlob [info script] [list */thatch.tcl]]]] > 0} {
			if 0 {
			foreach c [chan pipe] n [list r w] {
				chan configure $c -encoding utf-8 -translation lf -buffersize 32768
				set $n $c
			}
			puts -nonewline $w [thatchExtract $::argv0 $fid]
			close $w
			lappend fns -rename thatch.tcl
			set fn @$r
			}
			if 1 {
			set c [file tempfile t thatch]
			chan configure $c -encoding utf-8 -translation lf
			puts -nonewline $c [thatchExtract $::argv0 $fid]
			close $c
			lappend fns -rename thatch.tcl
			set fn $t
			cfg+ toDel $t
			}
		} elseif {[info exists ::env(HOME)]} {
			set tfn $::env(HOME)/tcl/thatch/library/thatch.tcl
			if {[file exists $tfn]} {
				lappend fns -home [file dirname $tfn]
				set fn [file tail $tfn]
			}
		}
		lappend fns $fn
	}
	return $fns
}

proc args {args} {
	set c [dict create]
	if {[llength $args] == 0} { dict set c usage 1; return $c }
	dict set c verbosity [cfg verbosity]
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-l - -c - -a - -r - -e - -y {
				if {[dict exists $c action]} {
					puts stderr "options -lcarey are mutually exclusive"; exit 1
				}
				dict set c action [string map [list -l list -c create -a add -r remove -e get -y copy] $arg]
			}
			-tb { dict set c thatchboot 1 }
			-wp { dict set c wishpreamble 1 }
			-np { dict set c preamble 0 }
			-v  { dict incr c verbosity }
			-pp { dict set c printpreamble 1 }
			-pb { dict set c printboot 1 }
			-o - -h - -H - -m - -to { set state $arg }
			-?  { dict set c help 1 }
			default {
				if {![dict exists $c inFile]} {
					dict set c inFile $arg
				} else {
					dict lappend c cliFiles $arg
				}
			}
			}
		}
		-o  { set state opt; dict set c outFile $arg }
		-h  { set state opt; set home $arg }
		-H  { set state opt; set Home $arg }
		-m  { set state opt; dict set c main $arg }
		-to { set state opt; dict set c thatchbootopts $arg }
		}; #switch
	}
	if {[dict exists $c action] && [dict exists $c gui] && [dict get $c gui]} {
		puts stderr "options -lcarey and -g are mutually exclusive"; exit 1
	}
	if {[info exists Home]} {
		set home $Home
		if {![dict exists $c cliFiles]} { dict set c cliFiles [list] }
		dict set c cliFiles [linsert [dict get $c cliFiles] 0 -home $home]
	}
	if {[info exists home]} {
		dict set c inFile [file join $home [dict get $c inFile]]
	}
	return $c
}

proc setup {} {
	::aitcs::aitcs+ new cfg
	eval cfg [join [lrange [split {
		NS			[namespace current]
		name			ThatchTool
		ver			0.5
		author			"Stuart Cassoff"
		when			"Summer 2022"
		msgs			{}
		projectFields		[dict create
						name        ""
						projectFile ""
						home        ""
						inFile      ""
						outFile     ""
						lastFileDir ""
						mtimeFormat ""
					]
		projectEditFields	[list
						name
						projectFile
						home
						inFile
						outFile
					]
		thatchFields		[list
						name
						size
						mtime
						base64
						zip
						md5
						sha256
						offset
						csize
					]
		thatchTvFields		[list
						name
						size
						mtime
						transforms
						checksums
						offset
						csize
					]
		mapCanvWidth		40
		mapCanvHeight		60
		mapCanvForeground	pink
		mapCanvBackground	darkgreen
		theme			default
		autoloadableProjects	[list]
		autoloadProjects	0
		addDefSize		0
		addDefMtime		0
		addDefBase64		0
		addDefZip		0
		addDefMd5		0
		addDefSha256		0
		configFile		thatchtool.conf
		autoloadConfigFile	1
		configFileFields	[list
						theme
						autoloadableProjects
						autoloadProjects
						addDefSize
						addDefMtime
						addDefBase64
						addDefZip
						addDefMd5
						addDefSha256
					]
		configBoolFields	[list
						autoloadProjects
						addDefSize
						addDefMtime
						addDefBase64
						addDefZip
						addDefMd5
						addDefSha256
					]
		filetypes 		[dict create
						tp	{{Thatch Project Files} *.tp}
						config	{{Thatch Configuration Files} .conf}
						all	{{All Files} *}
					]
		gui			no
		usage			0
		help			0
		verbosity		1
		action			list
		inFile			""
		outFile			"-"
		cliFiles		{}
		toDel			{}
		printpreamble		0
		printboot		0
		main			""
		thatchboot		0
		thatchbootopts		""
		wishpreamble		0
		preamble		1
	} \n] 1 end-1] " \\\n"]

	cfg fullname [cfg name]\ [cfg ver]

	interp alias {} pjs! {} variable pjs
	pjs!
	set pjs [dict create]
}

proc setupGui {} {
	package require central::extras
	package require gub
	namespace import ::gub::*
	package require meb
	namespace import ::meb::meb
	package require tu::toplevels::tk
	namespace import ::tu::toplevels::tk::*
	package require tu::epm
	namespace import ::tu::epm::*
	package require tu::pretty
	namespace import ::tu::pretty::hexdump
	package require Tk
	package require autoscroll
	::aitcs::aitcs new gui
	createGui
}


#---------------------------------------------------------------------
# Go
#---------------------------------------------------------------------

proc go {} {
	package require Tcl 8.6
	package require thatch
	namespace import ::thatch::*
	package require aitcs
	namespace import ::aitcs::aitcs
	package require tu
	package require tu::file
	namespace import ::tu::*
	namespace import ::tu::file::*
	setup
	cfg {*}[args {*}$::argv]; set ::argv {}; set ::argc 0
#	if {[info exists [cfg file]]} { set cfg($cfg(what)file) $cfg(file) }
	if {[cfg usage]}         { puts [usage]; return }
	if {[cfg help]}          { puts [usage]\n[help]; if {[cfg verbosity] > 1} { puts \nREADME:\n[readme] }; if {[cfg verbosity] > 2} { puts \nLICENSE:\n[join [dict values [LICENSE]] \n]}; return }
	if {[cfg printpreamble]} { puts [preamble] }
	if {[cfg printboot]}     { puts [string map [list @VER@ [package provide thatch]] $::thatch::bootstrap] }
	if {[cfg printpreamble] || [cfg printboot]} { return }
	autoloadConfig
	cfg cliFiles [resolveThatch [cfg cliFiles]]
	if {[cfg gui]} {
		setupGui
		autoloadProjects
		if {[cfg inFile] ne ""} {
			if {[file extension [cfg inFile]] eq ".tp"} {
				projectOpen [cfg inFile]
			} else {
				openThatch [cfg inFile]
			}
		}
	} else {
		cgogo
	}
}

namespace current}]::go

# EOF

# THATCH /pkgs/dsc/aitcs-2.0b1.tm size 9704 mtime 1660225868 modbase /pkgs/dsc
#
# AitCS - Array in the Command's Shadow
#
# Stuart Cassoff
#
#
# Spring 2018
# Version 1.1
#
# Winter 2020
# Version 1.2
#
# Spring 2020
# Version 1.3, 1.4, 1.5, 1.6, 1.7, 1.8
#
# Fall 2020
# Version 1.9
# Version 2.0b1
#


if {$::tcl_platform(engine) eq "Tcl"} { package require Tcl 8.6 }

namespace eval aitcs {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	aitcs
		title	AitCS
		descr	{Array in the Command's Shadow}
		author	{Stuart Cassoff}
		when	{Fall 2020}
		version	2.0b1
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}


## Internal

if {$::tcl_platform(engine) eq "Jim"} {
	proc Alias {this that args} {        alias    $this    $that {*}$args }
} else {
	proc Alias {this that args} { interp alias {} $this {} $that {*}$args }
}

proc Randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

proc Randvar {name variable_namespace} {
	return [string trimright $variable_namespace :]::__${name}__[Randstr]_
}

proc MakeCommands {type name varname cns} {
	set jim [expr {$::tcl_platform(engine) eq "Jim"}]

	set cns [string trimright $cns :]
	set smae [expr {[namespace qualifiers $varname] eq $cns}]
	set cns $cns:

	set ops   [list ""   ?       -      ^    &    ->    ]
	set procs [list Vset Vexists Vunset Vgwd Vref Vaitcs]

	if {$type eq "eagle"} {
		lappend ops   =      +        +=
		lappend procs Vsetto Vlappend Vincr
	}

	if {$jim} {
		set hows [lrepeat [llength $ops] full]
	} else {
		if {$smae} {
			set hows [lrepeat [llength $ops] var]
		} else {
			set hows [lrepeat [llength $ops] full]
			set pl [list Vexists Vgwd]
			if {$type eq "eagle"} {
				lappend pl Vsetto
			}
			foreach p $pl {
				lset hows [lsearch $procs $p] partial
			}
		}
		lset hows [lsearch $procs Vref] full
	}

	foreach op $ops proc $procs how $hows {
		set pargs [info args $proc]
		if {!$jim} {
			set pargs [lmap arg $pargs {
				expr {[info default $proc $arg def] ? [list $arg $def] : $arg}
			}]
		}
		set body [info body $proc]
		if {[lindex $pargs 0] eq "varname"} {
			set pargs [lrange $pargs 1 end]
			# For itiming tests, consult Makefile.aitcs.in when modifying this line: set how ?
			set body [switch -exact -- $how {
				var { # Change "upvar ..." to "variable ..." and hardcode [namespace tail $varname] everywhere
					set vn [namespace tail $varname]
					string map [list {upvar 1 $varname _v_} "variable $vn" _v_ $vn {$varname} $vn {${varname}} $vn] $body
				}
				full { # Remove "upvar ..." and hardcode $varname everywhere
					string map [list {upvar 1 $varname _v_} {} _v_ $varname {$varname} $varname {${varname}} $varname] $body
				}
				partial { # Hardcode $varname where used
					string map [list {$varname} $varname {${varname}} $varname] $body
				}
				setvar { # Add "set varname $varname" to top of proc
					format {set varname {%s}%s%s} $varname \n $body
				}
				alias { # Use aliases
					if {$proc eq "Vaitcs"} {
						if {[llength [info commands Vaitcs_alias]] == 0} {
							proc Vaitcs_alias {varname elem args} [string map [list \
								{tailcall $_v_($elem)$op {*}$args} \
								{uplevel 1 $_v_($elem)$op $args}] $body]
						}
						#set proc [string cat $proc _alias]
						append proc _alias
					}
					Alias $cns:$name$op [namespace current]::$proc $varname
					continue
				}
			}]
		}

		# Debug
		#puts "$name$op {$pargs} {$body\n}"

		proc $cns:$name$op $pargs $body

		# Debug
		#puts $name$op\n[::tcl::unsupported::disassemble proc $cns:$name$op]
	}

	# Debug
	#exit

	Alias $cns:$name! upvar #0 $varname $name

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name&&    return -level 0 $cns:$name
	proc    $cns:$name&& {} [list return $cns:$name]

	set --- [join [list {} \
		"unset $varname" \
		{*}[lmap op [list {*}$ops ! && ---] {
			format {rename %s%-3s {}} $cns:$name $op
		}] \
	] \n\t]\n

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name--- eval ${---}
	proc    $cns:$name--- {}   ${---}
}

proc New {type args} {
	set name ""
	set cns ""
	set vns ""
	set argies [list]
	set nameset no
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-cns - -command_namespace - -vns - -variable_namespace { set state $arg }
			-- { set state -- }
			default { if {$nameset} { lappend argies $arg } else { set name $arg; set nameset yes } }
			}; # hctiws
		}
		-cns - -command_namespace  { set state opt; set cns $arg }
		-vns - -variable_namespace { set state opt; set vns $arg }
		-- { lappend argies $arg }
		default { error "Huh? <[expr {int(rand()*100000)}]>" }
		}; # hctiws
	}
	if {$state ni [list opt --]} { return -code error "no argument given for option \"$state\"" }
	if {[expr {[llength $argies] % 2}] != 0} { return -code error "initial values must be \"key value ...\"" }
	if {$vns eq ""} { set vns [uplevel 1 {namespace current}] }
	if {$cns eq ""} { set cns $vns }
	set vns [string trimright $vns :]
	set cns [string trimright $cns :]
	if {$name eq ""} { for {*}{
		{ set name [Randstr]; set i 0 }
		{ $i < 1000 && [llength [info commands ${cns}::$name]] > 0 }
		{ set name [Randstr]; incr i }
		{}
	}}
	if {[llength [info commands ${cns}::$name]] > 0} { return -code error "${cns}::$name already exists" }
	set varname [Randvar $name $vns]
	MakeCommands $type $name $varname $cns
	array set $varname $argies
	return ${cns}::$name
}


## Operators

# AitCS / AitCS+

#
proc Vset {varname elem args} {
	upvar 1 $varname _v_
	if {[llength $args] == 0} {
		set _v_($elem)
	} elseif {[llength $args] == 1} {
		set _v_($elem) [lindex $args 0]
	} else {
		array set _v_ [list $elem {*}$args]
		set _v_($elem)
	}
}

# -
proc Vunset {varname elem args} {
	upvar 1 $varname _v_
	foreach e $args { array unset _v_ $e }
	array unset _v_ $elem
}

# &
proc Vref {varname {elem ""}} {
	if {$elem eq ""} { return $varname }
	return ${varname}($elem)
}

# ?
proc Vexists {varname elem {myVar ""}} {
	upvar 1 $varname _v_
	if {![info exists _v_($elem)]} { return 0 }
	if {$myVar eq {}} { return 1 }
	upvar 1 $myVar var
	set var $_v_($elem)
	return 1
}

# ^
proc Vgwd {varname elem {default ""} {myVar ""}} {
	upvar 1 $varname _v_
	if {$myVar ne ""} { upvar 1 $myVar var }
	set var [expr {[info exists _v_($elem)] ? $_v_($elem) : $default}]
}

# ->
proc Vaitcs {varname elem args} {
	upvar 1 $varname _v_
	# Avoid mismatches by checking double-char operators before single-char operators
	if {[set op [string range $elem end-1 end]] in [list += && ->]} {
		set elem [string range $elem 0 end-2]
	} elseif {[set op [string index $elem end]] in [list ? - ^ & = + ! :]} {
		set elem [string range $elem 0 end-1]
	} elseif {[set op [string range $elem end-2 end]] eq "---"} {
		set elem [string range $elem 0 end-3]
	} else {
		set op ""
	}
	tailcall $_v_($elem)$op {*}$args
}

# AitCS+

# =
proc Vsetto {varname elem elemOrVal args} {
	upvar 1 $varname _v_
	set a [list $elem $elemOrVal {*}$args]
	set val [lindex $a end]
	foreach e [lrange $a 0 end-1] { set _v_($e) $val }
	return $val
}

# +
proc Vlappend {varname elem args} {
	upvar 1 $varname _v_
	lappend _v_($elem) {*}$args
}

# +=
proc Vincr {varname elem {incr {1}} args} {
	upvar 1 $varname _v_
	foreach i $args { incr _v_($elem) $i }
	incr _v_($elem) $incr
}


## Setup

if {$::tcl_platform(engine) eq "Jim"} {
	foreach plus [list "" +] type [list snake eagle] {
		proc aitcs$plus {cmd args} [string map [list @plus@ $plus @type@ $type] {
			if {$cmd eq "operators"} {
				tailcall ::aitcs::$cmd@plus@ {*}$args
			} elseif {$cmd eq "new"} {
				tailcall ::aitcs::New @type@ {*}$args
			} else {
				return -code error "unknown subcommand \"$cmd\": must be new, or operators"
			}
		}]
	}
} else {
	foreach plus [list "" +] type [list snake eagle] {
		namespace ensemble create -prefixes no \
			-command aitcs$plus \
			-map [dict create new [list New $type] operators operators$plus]
	}
}


# These things
proc operators  {} { return [list "" ? - ^ & -> ! && ---] }
proc operators+ {} { return [list {*}[operators] = + +=] }


namespace export aitcs aitcs+

package provide aitcs 2.0b1


}; # namespace eval aitcs


# EOF

# THATCH /pkgs/dsc/tu-0.3.tm size 6680 mtime 1662628341 modbase /pkgs/dsc
#
# Tu - Tcl Usefulness
#
#
# Stuart Cassoff
#
# Fall/Winter 2018
# Version 0.1
#
# Summer 2020
# Version 0.2
#
# Summer 2022
# Version 0.3
#

package require Tcl 8.5

namespace eval tu {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu
		title	Tu
		descr	{Tcl Usefulness}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.3
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2017,2018,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


# Tcl

proc randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

# dict get with default
#
proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
#
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

#
# String Cat
#
proc strcat {args} { join $args "" }

#
# [lmap] replacement if not available
#
if {[llength [info commands lmap]] == 0} { proc lmap args {
	# 8.5 doesn't have lmap
	# https://wiki.tcl-lang.org/page/lmap+forward+compatibility
	set body [lindex $args end]
	set args [lrange $args 0 end-1]
	set n 0
	set pairs [list]
	foreach {varnames listval} $args {
		set varlist [list]
		foreach varname $varnames {
			upvar 1 $varname var$n
			lappend varlist var$n
			incr n
		}
		lappend pairs $varlist $listval
	}
	set temp [list]
	foreach {*}$pairs {
		lappend temp [uplevel 1 $body]
	}
	set temp
}}; # lmap

proc getNextOrPrevMatch {aList aString {curIdx 0} {backwards no}} {
	set idxs [lsearch -all -nocase $aList $aString*]
	if {[llength $idxs] == 0} { return -1 }
	if {[llength $idxs] == 1 && [lindex $idxs 0] == $curIdx} { return -1 }
	set p -1
	foreach idx $idxs {
		if {$idx < $curIdx} { set p $idx }
		if {$idx > $curIdx} { break }
	}
	if {$backwards} {
		if {$p == -1} {
			set newIdx [lindex $idxs end]
		} else {
			set newIdx $p
		}	
	} else {
		if {$idx == $curIdx} {
			set newIdx [lindex $idxs 0]
		} else {
			set newIdx $idx
		}
	}
	return $newIdx
}

proc makeRWAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " args " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			if {[llength $args] == 1} { set @m@ [lindex $args 0] }
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}
proc makeROAccessors {args} {
	set c ""
	foreach m $args { append c method " " $m " " "{}" " {" \
		[string trimright [regsub -all -line {^ +} [string map [list @m@ $m] {
			return $@m@
		}] "\t"]] "\n}\n"
	}
	return [string trimright $c]
}

#
# Remove blank lines and comment lines, add space-backslash to all lines but the last.
# If a line contains ";#" (spaces ok too) then chares from those to the end of the line will be removed
# Useful for making big lists in code without all the unpleasant backslashes.
#
proc backslashify {blob} { return [join [lmap l [split $blob \n] { if {[regexp {^[[:space:]]*(;?#.*)?$} $l]} { continue } else { lindex [regexp -inline -- {^(.+?)(;\s*#.*)?$} $l] 1 } }] " \\\n"] }

proc maxstrlen {args} { return [::tcl::mathfunc::max 0 {*}[lmap s $args { string length $s }]] }

proc oneOf {l} {
	if {[llength $l] < 2} { return [join $l] }
	return [strcat "one of " [join [lrange $l 0 end-1] ", "] [expr {[llength $l] > 2 ? "," : ""}] " or " [lindex $l end]]
}


## Tk

proc bindings {tag} {
	set d [dict create]
	set l [expr {[string index $tag 0] eq "." ? [bindtags $tag] : [list $tag]}]
	foreach t $l { foreach b [bind $t] {
		dict set d $t $b [bind $t $b]
	}}
	return $d
}

proc setEntryText {w text} {
	$w delete 0 end
	if {$text ne ""} { $w insert end $text }
}

proc setTextText {w text} {
	if {[set state [$w cget -state]] ne "normal"} { $w configure -state normal }
	$w delete 1.0 end
	if {$text ne ""} { $w insert end $text }
	#if {$text ne ""} { $w insert end $text[expr {[string index $text end] eq "\n" ? "" : "\n"}] }
	if {$state ne "normal"} { $w configure -state $state }
}

# From the Tk widget demo "A multi-column list of countries"
# Code to do the sorting of the tree contents when clicked on
#
proc tvSortBy {tree col direction} {
	# Build something we can sort
	set data {}
	foreach row [$tree children {}] {
		lappend data [list [$tree set $row $col] $row]
	}

	set dir [expr {$direction ? "-decreasing" : "-increasing"}]
	set r -1

	# Now reshuffle the rows into the sorted order
	foreach info [lsort -dictionary -index 0 $dir $data] {
		$tree move [lindex $info 1] {} [incr r]
	}

	# Switch the heading so that it will sort in the opposite direction
	$tree heading $col -command [list [namespace current]::tvSortBy $tree $col [expr {!$direction}]]
}

#
# Underline Ampersand
# Convert ampersand in text to -underline option
# From Gub
#
proc ua {text} {	
	set index 0
	set uli -1
	while {[set index [string first "&" $text $index]] != -1} {
		if {[string index $text $index+1] eq "&"} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			incr index
			continue
		}
		if {$index < [string length $text]-1} {
			set text [string range $text 0 $index-1][string range $text $index+1 end]
			set uli $index
		}
		incr index
	}
	set res [list $text]
	if {$uli != -1} {
		lappend res -underline $uli
	}
	return $res
}


## Setup

namespace export makeRWAccessors makeRWOccessors
namespace export randstr dgwd imn backslashify maxstrlen oneOf
namespace export tvSortBy setEntryText setTextText ua bindings

package provide tu 0.3


}; # namespace eval tu

# EOF

# THATCH /pkgs/dsc/tu/file-0.1.tm size 5076 mtime 1662579104 modbase /pkgs/dsc
#
# Tu File - Tcl Usefulness File Utils
#
#
# Stuart Cassoff
#
# Summer 2022
# Version 0.1
#

package require Tcl 8.5

namespace eval tu { namespace eval file {

namespace eval pkgconfig {
	variable pkgconfig [dict create {*}{
		name	tu::file
		title	{Tu File}
		descr	{File Utils}
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.1
	}]
	dict set pkgconfig importfrom [namespace parent]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	proc list {} { variable pkgconfig; return [dict keys $pkgconfig] }
	proc get {key} {
		variable pkgconfig
		if {[dict exists $pkgconfig $key]} {
			return [dict get $pkgconfig $key]
		} else {
			return -code error "key \"$key\" not known"
		}
	}
	namespace export get list
	namespace ensemble create
}; # pkgconfig


proc getFileFor {what {fn ""} args} {
	if {$fn ni {"" "?"}} { return $fn }
	if {[package provide Tk] eq ""} { return "" }
	return [${what}FileDialog {*}$args]
}

proc getLoadFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor load $fn {*}$args]
}

proc getSaveFile {{fn ""} args} {
	dict lappend args -filetypes [list "All Files" [list *]]
	return [getFileFor save $fn {*}$args]
}

proc openFile {fn {mode r}} {
	if {$fn eq "-"} { return [expr {$mode eq "r" ? "stdin" : "stdout"}] }
	if {$fn in {stdin stdout stderr}} { return $fn }
	if {[catch {open $fn $mode} f]} { return "" }
	return $f
}

proc closeFile {f} {
	if {$f in {stdin stdout stderr}} { return }
	close $f
}

proc loadFile {{fn ""} args} {
	set fn [getLoadFile $fn {*}$args]
	if {$fn eq ""} { return [list "" "" 0] }
	if {[set f [openFile $fn]] eq ""} { return [list "" "" 0] }
	set data [read -nonewline $f]
	closeFile $f
	return [list $fn $data 1]
}

proc saveFile {fn data args} {
	set fn [getSaveFile $fn {*}$args]
	if {$fn eq ""} { return [list "" 0] }
	if {[set f [openFile $fn w]] eq ""} { return [list "" "" 0] }
	puts -nonewline $f $data
	closeFile $f
	return [list $fn 1]
}

proc loadListFile {{fn ""} args} {
	if {![lassign [loadFile $fn {*}$args] fn data]} { return [list $fn "" 0] }
	return [list $fn [processListFileBlob $data] 1]
}

proc processListFileBlob {blob} {
	set res [list]
	foreach l [split $blob \n] {
		set l [string trim $l]
		if {$l eq {} || [string index $l 0] eq {#}} { continue }
		if {[catch {set len [llength $l]}] || $len < 1} { continue }
		lappend res $l
	}
	return $res
}

proc filemode->text {mode} {
    foreach { mask pairs } {
	00400 { 00400 r }
	00200 { 00200 w }
	04100 { 04100 s 04000 S 00100 x }
	00040 { 00040 r }
	00020 { 00020 w }
	02010 { 02010 s 02000 S 00010 x }
	00004 { 00004 r }
	00002 { 00002 w }
	01001 { 01001 t 01000 T 00001 x }
    } {
	set value [expr $mode & $mask]
	set bit -
	foreach { x b } $pairs {
	    if { $value == $x } {
		set bit $b
	    }
	}
	append retval $bit
    }
    return $retval
}


## Tk

proc loadFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getOpenFile {*}$args
	} else {
		tailcall ttk::getOpenFile {*}$args
	}
}

proc saveFileDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_getSaveFile {*}$args
	} else {
		tailcall ttk::getSaveFile {*}$args
	}
}

proc chooseDirectoryDialog {args} {
	if {$::tcl_platform(platform) ne "unix" || [catch {package require fsdialog}]} {
		tailcall tk_chooseDirectory {*}$args
	} else {
		tailcall ttk::chooseDirectory {*}$args
	}
}

proc loadFileIntoText {w {fn ""} args} {
	lassign [loadFile $fn {*}$args] fn data
	if {$fn eq ""} { return "" }
	$w delete 1.0 end
	$w insert end $data
	return $fn
}

proc saveSelectionToFile {w {fn ""} args} {
	if {[llength [set sel [$w tag ranges sel]]] != 2} { return }
	set blob [string trim [$w get {*}$sel]]
	if {$blob eq ""} { return }
	saveFile $fn $blob {*}$args
	return $fn
}


## Setup

package require tu

namespace path [namespace parent]

namespace export openFile closeFile getLoadFile loadFile getSaveFile saveFile
namespace export loadListFile processListFileBlob
namespace export loadFileDialog saveFileDialog
namespace export loadFileIntoText saveSelectionToFile
namespace export filemode->text

package provide tu::file 0.1

}}; # namespace tu::file


# EOF

# THATCH /pkgs/tcllib/md5/pkgIndex.tcl size 195 mtime 1659183356
if {![package vsatisfies [package provide Tcl] 8.2]} {return}
package ifneeded md5 2.0.8 [list source [file join $dir md5x.tcl]]
package ifneeded md5 1.4.5 [list source [file join $dir md5.tcl]]

# THATCH /pkgs/tcllib/md5/md5x.tcl size 25283 mtime 1659183356
# md5.tcl - Copyright (C) 2003 Pat Thoyts <patthoyts@users.sourceforge.net>
#
# MD5  defined by RFC 1321, "The MD5 Message-Digest Algorithm"
# HMAC defined by RFC 2104, "Keyed-Hashing for Message Authentication"
#
# This is an implementation of MD5 based upon the example code given in
# RFC 1321 and upon the tcllib MD4 implementation and taking some ideas
# from the earlier tcllib md5 version by Don Libes.
#
# This implementation permits incremental updating of the hash and 
# provides support for external compiled implementations either using
# critcl (md5c) or Trf.
#
# -------------------------------------------------------------------------
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# -------------------------------------------------------------------------

package require Tcl 8.2;                # tcl minimum version

namespace eval ::md5 {
    variable  accel
    array set accel {critcl 0 cryptkit 0 trf 0}

    namespace export md5 hmac MD5Init MD5Update MD5Final

    variable uid
    if {![info exists uid]} {
        set uid 0
    }
}

# -------------------------------------------------------------------------

# MD5Init --
#
#   Create and initialize an MD5 state variable. This will be
#   cleaned up when we call MD5Final
#
proc ::md5::MD5Init {} {
    variable accel
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token state

    # RFC1321:3.3 - Initialize MD5 state structure
    array set state \
        [list \
             A [expr {0x67452301}] \
             B [expr {0xefcdab89}] \
             C [expr {0x98badcfe}] \
             D [expr {0x10325476}] \
             n 0 i "" ]
    if {$accel(cryptkit)} {
        cryptkit::cryptCreateContext state(ckctx) CRYPT_UNUSED CRYPT_ALGO_MD5
    } elseif {$accel(trf)} {
        set s {}
        switch -exact -- $::tcl_platform(platform) {
            windows { set s [open NUL w] }
            unix    { set s [open /dev/null w] }
        }
        if {$s != {}} {
            fconfigure $s -translation binary -buffering none
            ::md5 -attach $s -mode write \
                -read-type variable \
                -read-destination [subst $token](trfread) \
                -write-type variable \
                -write-destination [subst $token](trfwrite)
            array set state [list trfread 0 trfwrite 0 trf $s]
        }
    }
    return $token
}

# MD5Update --
#
#   This is called to add more data into the hash. You may call this
#   as many times as you require. Note that passing in "ABC" is equivalent
#   to passing these letters in as separate calls -- hence this proc 
#   permits hashing of chunked data
#
#   If we have a C-based implementation available, then we will use
#   it here in preference to the pure-Tcl implementation.
#
proc ::md5::MD5Update {token data} {
    variable accel
    upvar #0 $token state

    if {$accel(critcl)} {
        if {[info exists state(md5c)]} {
            set state(md5c) [md5c $data $state(md5c)]
        } else {
            set state(md5c) [md5c $data]
        }
        return
    } elseif {[info exists state(ckctx)]} {
        if {[string length $data] > 0} {
            cryptkit::cryptEncrypt $state(ckctx) $data
        }
        return
    } elseif {[info exists state(trf)]} {
        puts -nonewline $state(trf) $data
        return
    }

    # Update the state values
    incr state(n) [string length $data]
    append state(i) $data

    # Calculate the hash for any complete blocks
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        MD5Hash $token [string range $state(i) $n [incr n 64]]
    }

    # Adjust the state for the blocks completed.
    set state(i) [string range $state(i) $n end]
    return
}

# MD5Final --
#
#    This procedure is used to close the current hash and returns the
#    hash data. Once this procedure has been called the hash context
#    is freed and cannot be used again.
#
#    Note that the output is 128 bits represented as binary data.
#
proc ::md5::MD5Final {token} {
    upvar #0 $token state

    # Check for either of the C-compiled versions.
    if {[info exists state(md5c)]} {
        set r $state(md5c)
        unset state
        return $r
    } elseif {[info exists state(ckctx)]} {
        cryptkit::cryptEncrypt $state(ckctx) ""
        cryptkit::cryptGetAttributeString $state(ckctx) \
            CRYPT_CTXINFO_HASHVALUE r 16
        cryptkit::cryptDestroyContext $state(ckctx)
        # If nothing was hashed, we get no r variable set!
        if {[info exists r]} {
            unset state
            return $r
        }
    } elseif {[info exists state(trf)]} {
        close $state(trf)
        set r $state(trfwrite)
        unset state
        return $r
    }

    # RFC1321:3.1 - Padding
    #
    set len [string length $state(i)]
    set pad [expr {56 - ($len % 64)}]
    if {$len % 64 > 56} {
        incr pad 64
    }
    if {$pad == 0} {
        incr pad 64
    }

    #puts "P $pad|bits=[expr {8 * $state(n)}]"

    append state(i) [binary format a$pad \x80]

    # RFC1321:3.2 - Append length in bits as little-endian wide int.
    append state(i) [binary format ii [expr {8 * $state(n)}] 0]

    #puts DATA=[Hex $state(i)]([string length $state(i)])

    # Calculate the hash for the remaining block.
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        MD5Hash $token [string range $state(i) $n [incr n 64]]
    }

    #puts md5-post__________________________________________
    #parray ::${token}

    # RFC1321:3.5 - Output
    set r [bytes $state(A)][bytes $state(B)][bytes $state(C)][bytes $state(D)]
    unset state

    #puts HASH=[Hex $r]
    return $r
}

# -------------------------------------------------------------------------
# HMAC Hashed Message Authentication (RFC 2104)
#
# hmac = H(K xor opad, H(K xor ipad, text))
#

# HMACInit --
#
#    This is equivalent to the MD5Init procedure except that a key is
#    added into the algorithm
#
proc ::md5::HMACInit {K} {

    # Key K is adjusted to be 64 bytes long. If K is larger, then use
    # the MD5 digest of K and pad this instead.
    set len [string length $K]
    if {$len > 64} {
        set tok [MD5Init]
        MD5Update $tok $K
        set K [MD5Final $tok]
        set len [string length $K]
    }
    set pad [expr {64 - $len}]
    append K [string repeat \0 $pad]

    # Cacluate the padding buffers.
    set Ki {}
    set Ko {}
    binary scan $K i16 Ks
    foreach k $Ks {
        append Ki [binary format i [expr {$k ^ 0x36363636}]]
        append Ko [binary format i [expr {$k ^ 0x5c5c5c5c}]]
    }

    set tok [MD5Init]
    MD5Update $tok $Ki;                 # initialize with the inner pad
    
    # preserve the Ko value for the final stage.
    # FRINK: nocheck
    set [subst $tok](Ko) $Ko

    return $tok
}

# HMACUpdate --
#
#    Identical to calling MD5Update
#
proc ::md5::HMACUpdate {token data} {
    MD5Update $token $data
    return
}

# HMACFinal --
#
#    This is equivalent to the MD5Final procedure. The hash context is
#    closed and the binary representation of the hash result is returned.
#
proc ::md5::HMACFinal {token} {
    upvar #0 $token state

    set tok [MD5Init];                  # init the outer hashing function
    MD5Update $tok $state(Ko);          # prepare with the outer pad.
    MD5Update $tok [MD5Final $token];   # hash the inner result
    return [MD5Final $tok]
}

# -------------------------------------------------------------------------
# Description:
#  This is the core MD5 algorithm. It is a lot like the MD4 algorithm but
#  includes an extra round and a set of constant modifiers throughout.
# 
# Note:
#  This function body is substituted later on to inline some of the 
#  procedures and to make is a bit more comprehensible.
#
set ::md5::MD5Hash_body {
    variable $token
    upvar 0 $token state

    #puts TR__=[Hex $msg]([string length $msg])

    # RFC1321:3.4 - Process Message in 16-Word Blocks
    binary scan $msg i* blocks
    foreach {X0 X1 X2 X3 X4 X5 X6 X7 X8 X9 X10 X11 X12 X13 X14 X15} $blocks {
        #puts BL

        set A $state(A)
        set B $state(B)
        set C $state(C)
        set D $state(D)

        # Round 1
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + F(b,c,d) + X[k] + T[i]) <<< s).
        # Do the following 16 operations.
        # [ABCD  0  7  1]  [DABC  1 12  2]  [CDAB  2 17  3]  [BCDA  3 22  4]
        set A [expr {$B + (($A + [F $B $C $D] + $X0 + $T01) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X1 + $T02) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X2 + $T03) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X3 + $T04) <<< 22)}]
        # [ABCD  4  7  5]  [DABC  5 12  6]  [CDAB  6 17  7]  [BCDA  7 22  8]
        set A [expr {$B + (($A + [F $B $C $D] + $X4 + $T05) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X5 + $T06) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X6 + $T07) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X7 + $T08) <<< 22)}]
        # [ABCD  8  7  9]  [DABC  9 12 10]  [CDAB 10 17 11]  [BCDA 11 22 12]
        set A [expr {$B + (($A + [F $B $C $D] + $X8 + $T09) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X9 + $T10) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X10 + $T11) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X11 + $T12) <<< 22)}]
        # [ABCD 12  7 13]  [DABC 13 12 14]  [CDAB 14 17 15]  [BCDA 15 22 16]
        set A [expr {$B + (($A + [F $B $C $D] + $X12 + $T13) <<< 7)}]
        set D [expr {$A + (($D + [F $A $B $C] + $X13 + $T14) <<< 12)}]
        set C [expr {$D + (($C + [F $D $A $B] + $X14 + $T15) <<< 17)}]
        set B [expr {$C + (($B + [F $C $D $A] + $X15 + $T16) <<< 22)}]

        # Round 2.
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + G(b,c,d) + X[k] + Ti) <<< s)
        # Do the following 16 operations.
        # [ABCD  1  5 17]  [DABC  6  9 18]  [CDAB 11 14 19]  [BCDA  0 20 20]
        set A [expr {$B + (($A + [G $B $C $D] + $X1  + $T17) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X6  + $T18) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X11 + $T19) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X0  + $T20) <<< 20)}]
        # [ABCD  5  5 21]  [DABC 10  9 22]  [CDAB 15 14 23]  [BCDA  4 20 24]
        set A [expr {$B + (($A + [G $B $C $D] + $X5  + $T21) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X10 + $T22) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X15 + $T23) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X4  + $T24) <<< 20)}]
        # [ABCD  9  5 25]  [DABC 14  9 26]  [CDAB  3 14 27]  [BCDA  8 20 28]
        set A [expr {$B + (($A + [G $B $C $D] + $X9  + $T25) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X14 + $T26) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X3  + $T27) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X8  + $T28) <<< 20)}]
        # [ABCD 13  5 29]  [DABC  2  9 30]  [CDAB  7 14 31]  [BCDA 12 20 32]
        set A [expr {$B + (($A + [G $B $C $D] + $X13 + $T29) <<<  5)}]
        set D [expr {$A + (($D + [G $A $B $C] + $X2  + $T30) <<<  9)}]
        set C [expr {$D + (($C + [G $D $A $B] + $X7  + $T31) <<< 14)}]
        set B [expr {$C + (($B + [G $C $D $A] + $X12 + $T32) <<< 20)}]
        
        # Round 3.
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + H(b,c,d) + X[k] + T[i]) <<< s)
        # Do the following 16 operations.
        # [ABCD  5  4 33]  [DABC  8 11 34]  [CDAB 11 16 35]  [BCDA 14 23 36]
        set A [expr {$B + (($A + [H $B $C $D] + $X5  + $T33) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X8  + $T34) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X11 + $T35) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X14 + $T36) <<< 23)}]
        # [ABCD  1  4 37]  [DABC  4 11 38]  [CDAB  7 16 39]  [BCDA 10 23 40]
        set A [expr {$B + (($A + [H $B $C $D] + $X1  + $T37) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X4  + $T38) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X7  + $T39) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X10 + $T40) <<< 23)}]
        # [ABCD 13  4 41]  [DABC  0 11 42]  [CDAB  3 16 43]  [BCDA  6 23 44]
        set A [expr {$B + (($A + [H $B $C $D] + $X13 + $T41) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X0  + $T42) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X3  + $T43) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X6  + $T44) <<< 23)}]
        # [ABCD  9  4 45]  [DABC 12 11 46]  [CDAB 15 16 47]  [BCDA  2 23 48]
        set A [expr {$B + (($A + [H $B $C $D] + $X9  + $T45) <<<  4)}]
        set D [expr {$A + (($D + [H $A $B $C] + $X12 + $T46) <<< 11)}]
        set C [expr {$D + (($C + [H $D $A $B] + $X15 + $T47) <<< 16)}]
        set B [expr {$C + (($B + [H $C $D $A] + $X2  + $T48) <<< 23)}]

        # Round 4.
        # Let [abcd k s i] denote the operation
        #   a = b + ((a + I(b,c,d) + X[k] + T[i]) <<< s)
        # Do the following 16 operations.
        # [ABCD  0  6 49]  [DABC  7 10 50]  [CDAB 14 15 51]  [BCDA  5 21 52]
        set A [expr {$B + (($A + [I $B $C $D] + $X0  + $T49) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X7  + $T50) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X14 + $T51) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X5  + $T52) <<< 21)}]
        # [ABCD 12  6 53]  [DABC  3 10 54]  [CDAB 10 15 55]  [BCDA  1 21 56]
        set A [expr {$B + (($A + [I $B $C $D] + $X12 + $T53) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X3  + $T54) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X10 + $T55) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X1  + $T56) <<< 21)}]
        # [ABCD  8  6 57]  [DABC 15 10 58]  [CDAB  6 15 59]  [BCDA 13 21 60]
        set A [expr {$B + (($A + [I $B $C $D] + $X8  + $T57) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X15 + $T58) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X6  + $T59) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X13 + $T60) <<< 21)}]
        # [ABCD  4  6 61]  [DABC 11 10 62]  [CDAB  2 15 63]  [BCDA  9 21 64]
        set A [expr {$B + (($A + [I $B $C $D] + $X4  + $T61) <<<  6)}]
        set D [expr {$A + (($D + [I $A $B $C] + $X11 + $T62) <<< 10)}]
        set C [expr {$D + (($C + [I $D $A $B] + $X2  + $T63) <<< 15)}]
        set B [expr {$C + (($B + [I $C $D $A] + $X9  + $T64) <<< 21)}]

        # Then perform the following additions. (That is, increment each
        # of the four registers by the value it had before this block
        # was started.)
        incr state(A) $A
        incr state(B) $B
        incr state(C) $C
        incr state(D) $D
    }

    return
}

proc ::md5::byte {n v} {expr {((0xFF << (8 * $n)) & $v) >> (8 * $n)}}
proc ::md5::bytes {v} { 
    #format %c%c%c%c [byte 0 $v] [byte 1 $v] [byte 2 $v] [byte 3 $v]
    format %c%c%c%c \
        [expr {0xFF & $v}] \
        [expr {(0xFF00 & $v) >> 8}] \
        [expr {(0xFF0000 & $v) >> 16}] \
        [expr {((0xFF000000 & $v) >> 24) & 0xFF}]
}

# 32bit rotate-left
proc ::md5::<<< {v n} {
    return [expr {((($v << $n) \
                        | (($v >> (32 - $n)) \
                               & (0x7FFFFFFF >> (31 - $n))))) \
                      & 0xFFFFFFFF}]
}

# Convert our <<< pseudo-operator into a procedure call.
regsub -all -line \
    {\[expr {(\$[ABCD]) \+ \(\((.*)\)\s+<<<\s+(\d+)\)}\]} \
    $::md5::MD5Hash_body \
    {[expr {int(\1 + [<<< [expr {\2}] \3])}]} \
    ::md5::MD5Hash_body

# RFC1321:3.4 - function F
proc ::md5::F {X Y Z} {
    return [expr {($X & $Y) | ((~$X) & $Z)}]
}

# Inline the F function
regsub -all -line \
    {\[F (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {( (\1 \& \2) | ((~\1) \& \3) )} \
    ::md5::MD5Hash_body
    
# RFC1321:3.4 - function G
proc ::md5::G {X Y Z} {
    return [expr {(($X & $Z) | ($Y & (~$Z)))}]
}

# Inline the G function
regsub -all -line \
    {\[G (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {(((\1 \& \3) | (\2 \& (~\3))))} \
    ::md5::MD5Hash_body

# RFC1321:3.4 - function H
proc ::md5::H {X Y Z} {
    return [expr {$X ^ $Y ^ $Z}]
}

# Inline the H function
regsub -all -line \
    {\[H (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {(\1 ^ \2 ^ \3)} \
    ::md5::MD5Hash_body

# RFC1321:3.4 - function I
proc ::md5::I {X Y Z} {
    return [expr {$Y ^ ($X | (~$Z))}]
}

# Inline the I function
regsub -all -line \
    {\[I (\$[ABCD]) (\$[ABCD]) (\$[ABCD])\]} \
    $::md5::MD5Hash_body \
    {(\2 ^ (\1 | (~\3)))} \
    ::md5::MD5Hash_body


# RFC 1321:3.4 step 4: inline the set of constant modifiers.
namespace eval md5 {
    variable tName
    variable tVal
    variable map
    foreach tName {
        T01 T02 T03 T04 T05 T06 T07 T08 T09 T10 
        T11 T12 T13 T14 T15 T16 T17 T18 T19 T20 
        T21 T22 T23 T24 T25 T26 T27 T28 T29 T30 
        T31 T32 T33 T34 T35 T36 T37 T38 T39 T40 
        T41 T42 T43 T44 T45 T46 T47 T48 T49 T50 
        T51 T52 T53 T54 T55 T56 T57 T58 T59 T60 
        T61 T62 T63 T64 
    }  tVal {
        0xd76aa478 0xe8c7b756 0x242070db 0xc1bdceee
        0xf57c0faf 0x4787c62a 0xa8304613 0xfd469501
        0x698098d8 0x8b44f7af 0xffff5bb1 0x895cd7be
        0x6b901122 0xfd987193 0xa679438e 0x49b40821
        
        0xf61e2562 0xc040b340 0x265e5a51 0xe9b6c7aa
        0xd62f105d 0x2441453  0xd8a1e681 0xe7d3fbc8
        0x21e1cde6 0xc33707d6 0xf4d50d87 0x455a14ed
        0xa9e3e905 0xfcefa3f8 0x676f02d9 0x8d2a4c8a
        
        0xfffa3942 0x8771f681 0x6d9d6122 0xfde5380c
        0xa4beea44 0x4bdecfa9 0xf6bb4b60 0xbebfbc70
        0x289b7ec6 0xeaa127fa 0xd4ef3085 0x4881d05
        0xd9d4d039 0xe6db99e5 0x1fa27cf8 0xc4ac5665
        
        0xf4292244 0x432aff97 0xab9423a7 0xfc93a039
        0x655b59c3 0x8f0ccc92 0xffeff47d 0x85845dd1
        0x6fa87e4f 0xfe2ce6e0 0xa3014314 0x4e0811a1
        0xf7537e82 0xbd3af235 0x2ad7d2bb 0xeb86d391
    } {
        lappend map \$$tName $tVal
    }
    set ::md5::MD5Hash_body [string map $map $::md5::MD5Hash_body]
    unset map tName tVal
}

# Define the MD5 hashing procedure with inline functions.
proc ::md5::MD5Hash {token msg} $::md5::MD5Hash_body
unset ::md5::MD5Hash_body

# -------------------------------------------------------------------------

if {[package provide Trf] != {}} {
    interp alias {} ::md5::Hex {} ::hex -mode encode --
} else {
    proc ::md5::Hex {data} {
        binary scan $data H* result
        return [string toupper $result]
    }
}

# -------------------------------------------------------------------------

# LoadAccelerator --
#
#	This package can make use of a number of compiled extensions to
#	accelerate the digest computation. This procedure manages the
#	use of these extensions within the package. During normal usage
#	this should not be called, but the test package manipulates the
#	list of enabled accelerators.
#
proc ::md5::LoadAccelerator {name} {
    variable accel
    set r 0
    switch -exact -- $name {
        critcl {
            if {![catch {package require tcllibc}]
                || ![catch {package require md5c}]} {
                set r [expr {[info commands ::md5::md5c] != {}}]
            }
        }
        cryptkit {
            if {![catch {package require cryptkit}]} {
                set r [expr {![catch {cryptkit::cryptInit}]}]
            }
        }
        trf {
            if {![catch {package require Trf}]} {
                set r [expr {![catch {::md5 aa} msg]}]
            }
        }
        default {
            return -code error "invalid accelerator package:\
                must be one of [join [array names accel] {, }]"
        }
    }
    set accel($name) $r
}

# -------------------------------------------------------------------------

# Description:
#  Pop the nth element off a list. Used in options processing.
#
proc ::md5::Pop {varname {nth 0}} {
    upvar $varname args
    set r [lindex $args $nth]
    set args [lreplace $args $nth $nth]
    return $r
}

# -------------------------------------------------------------------------

# fileevent handler for chunked file hashing.
#
proc ::md5::Chunk {token channel {chunksize 4096}} {
    upvar #0 $token state
    
    if {[eof $channel]} {
        fileevent $channel readable {}
        set state(reading) 0
    }
        
    MD5Update $token [read $channel $chunksize]
}

# -------------------------------------------------------------------------

proc ::md5::md5 {args} {
    array set opts {-hex 0 -filename {} -channel {} -chunksize 4096}
    while {[string match -* [set option [lindex $args 0]]]} {
        switch -glob -- $option {
            -hex       { set opts(-hex) 1 }
            -file*     { set opts(-filename) [Pop args 1] }
            -channel   { set opts(-channel) [Pop args 1] }
            -chunksize { set opts(-chunksize) [Pop args 1] }
            default {
                if {[llength $args] == 1} { break }
                if {[string compare $option "--"] == 0} { Pop args; break }
                set err [join [lsort [array names opts]] ", "]
                return -code error "bad option $option:\
                    must be one of $err\nlen: [llength $args]"
            }
        }
        Pop args
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {

        if {[llength $args] != 1} {
            return -code error "wrong # args:\
                should be \"md5 ?-hex? -filename file | string\""
        }
        set tok [MD5Init]

        #puts md5_______________________________________________
        #parray ::${tok}

        #puts IN=(([lindex $args 0]))
        MD5Update $tok [lindex $args 0]

        #puts md5-final_________________________________________
        #parray ::${tok}

        set r [MD5Final $tok]

    } else {

        set tok [MD5Init]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        vwait [subst $tok](reading)
        set r [MD5Final $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

# -------------------------------------------------------------------------

proc ::md5::hmac {args} {
    array set opts {-hex 0 -filename {} -channel {} -chunksize 4096}
    while {[string match -* [set option [lindex $args 0]]]} {
        switch -glob -- $option {
            -key       { set opts(-key) [Pop args 1] }
            -hex       { set opts(-hex) 1 }
            -file*     { set opts(-filename) [Pop args 1] }
            -channel   { set opts(-channel) [Pop args 1] }
            -chunksize { set opts(-chunksize) [Pop args 1] }
            default {
                if {[llength $args] == 1} { break }
                if {[string compare $option "--"] == 0} { Pop args; break }
                set err [join [lsort [array names opts]] ", "]
                return -code error "bad option $option:\
                    must be one of $err"
            }
        }
        Pop args
    }

    if {![info exists opts(-key)]} {
        return -code error "wrong # args:\
            should be \"hmac ?-hex? -key key -filename file | string\""
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {

        if {[llength $args] != 1} {
            return -code error "wrong # args:\
                should be \"hmac ?-hex? -key key -filename file | string\""
        }
        set tok [HMACInit $opts(-key)]
        HMACUpdate $tok [lindex $args 0]
        set r [HMACFinal $tok]

    } else {

        set tok [HMACInit $opts(-key)]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        vwait [subst $tok](reading)
        set r [HMACFinal $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

# -------------------------------------------------------------------------

# Try and load a compiled extension to help.
namespace eval ::md5 {
    variable e
    foreach  e {critcl cryptkit trf} { if {[LoadAccelerator $e]} { break } }
    unset    e
}

package provide md5 2.0.8

# -------------------------------------------------------------------------
# Local Variables:
#   mode: tcl
#   indent-tabs-mode: nil
# End:



# THATCH /pkgs/tcllib/sha1/pkgIndex.tcl size 816 mtime 1659183357
# Tcl package index file, version 1.1
# This file is generated by the "pkg_mkIndex" command
# and sourced either when an application starts up or
# by a "package unknown" script.  It invokes the
# "package ifneeded" command to set up package-related
# information so that packages will be loaded automatically
# in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the
# full path name of this file's directory.

if {![package vsatisfies [package provide Tcl] 8.2]} {return}
package ifneeded sha256 1.0.4 [list source [file join $dir sha256.tcl]]
package ifneeded sha256c 1.0.4 [list source [file join $dir sha256c.tcl]]
package ifneeded sha1   2.0.4 [list source [file join $dir sha1.tcl]]
package ifneeded sha1   1.1.1 [list source [file join $dir sha1v1.tcl]]

# THATCH /pkgs/tcllib/sha1/sha256.tcl size 24593 mtime 1659183357
# sha256.tcl - Copyright (C) 2005 Pat Thoyts <patthoyts@users.sourceforge.net>
#
# SHA1 defined by FIPS 180-2, "The Secure Hash Standard"
# HMAC defined by RFC 2104, "Keyed-Hashing for Message Authentication"
#
# This is an implementation of the secure hash algorithms specified in the
# FIPS 180-2 document.
#
# This implementation permits incremental updating of the hash and 
# provides support for external compiled implementations using critcl.
#
# This implementation permits incremental updating of the hash and 
# provides support for external compiled implementations either using
# critcl (sha256c).
#
# Ref: http://csrc.nist.gov/publications/fips/fips180-2/fips180-2.pdf
#      http://csrc.nist.gov/publications/fips/fips180-2/fips180-2withchangenotice.pdf
#
# -------------------------------------------------------------------------
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# -------------------------------------------------------------------------
# @mdgen EXCLUDE: sha256c.tcl

package require Tcl 8.2;                # tcl minimum version

namespace eval ::sha2 {
    variable  accel
    array set accel {tcl 0 critcl 0}
    variable  loaded {}

    namespace export sha256 hmac \
            SHA256Init SHA256Update SHA256Final


    variable uid
    if {![info exists uid]} {
        set uid 0
    }

    variable K
    if {![info exists K]} {
        # FIPS 180-2: 4.2.2 SHA-256 constants
        set K [list \
                   0x428a2f98 0x71374491 0xb5c0fbcf 0xe9b5dba5 \
                   0x3956c25b 0x59f111f1 0x923f82a4 0xab1c5ed5 \
                   0xd807aa98 0x12835b01 0x243185be 0x550c7dc3 \
                   0x72be5d74 0x80deb1fe 0x9bdc06a7 0xc19bf174 \
                   0xe49b69c1 0xefbe4786 0x0fc19dc6 0x240ca1cc \
                   0x2de92c6f 0x4a7484aa 0x5cb0a9dc 0x76f988da \
                   0x983e5152 0xa831c66d 0xb00327c8 0xbf597fc7 \
                   0xc6e00bf3 0xd5a79147 0x06ca6351 0x14292967 \
                   0x27b70a85 0x2e1b2138 0x4d2c6dfc 0x53380d13 \
                   0x650a7354 0x766a0abb 0x81c2c92e 0x92722c85 \
                   0xa2bfe8a1 0xa81a664b 0xc24b8b70 0xc76c51a3 \
                   0xd192e819 0xd6990624 0xf40e3585 0x106aa070 \
                   0x19a4c116 0x1e376c08 0x2748774c 0x34b0bcb5 \
                   0x391c0cb3 0x4ed8aa4a 0x5b9cca4f 0x682e6ff3 \
                   0x748f82ee 0x78a5636f 0x84c87814 0x8cc70208 \
                   0x90befffa 0xa4506ceb 0xbef9a3f7 0xc67178f2 \
                  ]
    }
    
}

# -------------------------------------------------------------------------
# Management of sha256 implementations.

# LoadAccelerator --
#
#	This package can make use of a number of compiled extensions to
#	accelerate the digest computation. This procedure manages the
#	use of these extensions within the package. During normal usage
#	this should not be called, but the test package manipulates the
#	list of enabled accelerators.
#
proc ::sha2::LoadAccelerator {name} {
    variable accel
    set r 0
    switch -exact -- $name {
        tcl {
            # Already present (this file)
            set r 1
        }
        critcl {
            if {![catch {package require tcllibc}]
                || ![catch {package require sha256c}]} {
                set r [expr {[info commands ::sha2::sha256c_update] != {}}]
            }
        }
        default {
            return -code error "invalid accelerator $key:\
                must be one of [join [KnownImplementations] {, }]"
        }
    }
    set accel($name) $r
    return $r
}

# ::sha2::Implementations --
#
#	Determines which implementations are
#	present, i.e. loaded.
#
# Arguments:
#	None.
#
# Results:
#	A list of implementation keys.

proc ::sha2::Implementations {} {
    variable accel
    set res {}
    foreach n [array names accel] {
	if {!$accel($n)} continue
	lappend res $n
    }
    return $res
}

# ::sha2::KnownImplementations --
#
#	Determines which implementations are known
#	as possible implementations.
#
# Arguments:
#	None.
#
# Results:
#	A list of implementation keys. In the order
#	of preference, most prefered first.

proc ::sha2::KnownImplementations {} {
    return {critcl tcl}
}

proc ::sha2::Names {} {
    return {
	critcl   {tcllibc based}
	tcl      {pure Tcl}
    }
}

# ::sha2::SwitchTo --
#
#	Activates a loaded named implementation.
#
# Arguments:
#	key	Name of the implementation to activate.
#
# Results:
#	None.

proc ::sha2::SwitchTo {key} {
    variable accel
    variable loaded

    if {[string equal $key $loaded]} {
	# No change, nothing to do.
	return
    } elseif {![string equal $key ""]} {
	# Validate the target implementation of the switch.

	if {![info exists accel($key)]} {
	    return -code error "Unable to activate unknown implementation \"$key\""
	} elseif {![info exists accel($key)] || !$accel($key)} {
	    return -code error "Unable to activate missing implementation \"$key\""
	}
    }

    # Deactivate the previous implementation, if there was any.

    if {![string equal $loaded ""]} {
        foreach c {
            SHA256Init   SHA224Init
            SHA256Final  SHA224Final
            SHA256Update
        } {
            interp alias {} ::sha2::$c {}
        }
    }

    # Activate the new implementation, if there is any.

    if {![string equal $key ""]} {
        foreach c {
            SHA256Init   SHA224Init
            SHA256Final  SHA224Final
            SHA256Update
        } {
	    interp alias {} ::sha2::$c {} ::sha2::${c}-${key}
        }
    }

    # Remember the active implementation, for deactivation by future
    # switches.

    set loaded $key
    return
}

# -------------------------------------------------------------------------

# SHA256Init --
#
#   Create and initialize an SHA256 state variable. This will be
#   cleaned up when we call SHA256Final
#

proc ::sha2::SHA256Init-tcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2: 5.3.2 Setting the initial hash value
    array set tok \
            [list \
            A [expr {int(0x6a09e667)}] \
            B [expr {int(0xbb67ae85)}] \
            C [expr {int(0x3c6ef372)}] \
            D [expr {int(0xa54ff53a)}] \
            E [expr {int(0x510e527f)}] \
            F [expr {int(0x9b05688c)}] \
            G [expr {int(0x1f83d9ab)}] \
            H [expr {int(0x5be0cd19)}] \
            n 0 i "" v 256]
    return $token
}

proc ::sha2::SHA256Init-critcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2: 5.3.2 Setting the initial hash value
    set tok(sha256c) [sha256c_init256]
    return $token
}

# SHA256Update --
#
#   This is called to add more data into the hash. You may call this
#   as many times as you require. Note that passing in "ABC" is equivalent
#   to passing these letters in as separate calls -- hence this proc 
#   permits hashing of chunked data
#
#   If we have a C-based implementation available, then we will use
#   it here in preference to the pure-Tcl implementation.
#

proc ::sha2::SHA256Update-tcl {token data} {
    upvar #0 $token state

    # Update the state values
    incr   state(n) [string length $data]
    append state(i) $data

    # Calculate the hash for any complete blocks
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        SHA256Transform $token [string range $state(i) $n [incr n 64]]
    }

    # Adjust the state for the blocks completed.
    set state(i) [string range $state(i) $n end]
    return
}

proc ::sha2::SHA256Update-critcl {token data} {
    upvar #0 $token state

    set state(sha256c) [sha256c_update $data $state(sha256c)]
    return
}

# SHA256Final --
#
#    This procedure is used to close the current hash and returns the
#    hash data. Once this procedure has been called the hash context
#    is freed and cannot be used again.
#
#    Note that the output is 256 bits represented as binary data.
#

proc ::sha2::SHA256Final-tcl {token} {
    upvar #0 $token state
    SHA256Penultimate $token
    
    # Output
    set r [bytes $state(A)][bytes $state(B)][bytes $state(C)][bytes $state(D)][bytes $state(E)][bytes $state(F)][bytes $state(G)][bytes $state(H)]
    unset state
    return $r
}

proc ::sha2::SHA256Final-critcl {token} {
    upvar #0 $token state
    set r $state(sha256c)
    unset  state
    return $r
}

# SHA256Penultimate --
#
#
proc ::sha2::SHA256Penultimate {token} {
    upvar #0 $token state

    # FIPS 180-2: 5.1.1: Padding the message
    #
    set len [string length $state(i)]
    set pad [expr {56 - ($len % 64)}]
    if {$len % 64 > 56} {
        incr pad 64
    }
    if {$pad == 0} {
        incr pad 64
    }
    append state(i) [binary format a$pad \x80]

    # Append length in bits as big-endian wide int.
    set dlen [expr {8 * $state(n)}]
    append state(i) [binary format II 0 $dlen]

    # Calculate the hash for the remaining block.
    set len [string length $state(i)]
    for {set n 0} {($n + 64) <= $len} {} {
        SHA256Transform $token [string range $state(i) $n [incr n 64]]
    }
}

# -------------------------------------------------------------------------

proc ::sha2::SHA224Init-tcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2 (change notice 1) (1): SHA-224 initialization values
    array set tok \
            [list \
            A [expr {int(0xc1059ed8)}] \
            B [expr {int(0x367cd507)}] \
            C [expr {int(0x3070dd17)}] \
            D [expr {int(0xf70e5939)}] \
            E [expr {int(0xffc00b31)}] \
            F [expr {int(0x68581511)}] \
            G [expr {int(0x64f98fa7)}] \
            H [expr {int(0xbefa4fa4)}] \
            n 0 i "" v 224]
    return $token
}

proc ::sha2::SHA224Init-critcl {} {
    variable uid
    set token [namespace current]::[incr uid]
    upvar #0 $token tok

    # FIPS 180-2 (change notice 1) (1): SHA-224 initialization values
    set tok(sha256c) [sha256c_init224]
    return $token
}

interp alias {} ::sha2::SHA224Update {} ::sha2::SHA256Update

proc ::sha2::SHA224Final-tcl {token} {
    upvar #0 $token state
    SHA256Penultimate $token
    
    # Output
    set r [bytes $state(A)][bytes $state(B)][bytes $state(C)][bytes $state(D)][bytes $state(E)][bytes $state(F)][bytes $state(G)]
    unset state
    return $r
}

proc ::sha2::SHA224Final-critcl {token} {
    upvar #0 $token state
    # Trim result down to 224 bits (by 4 bytes).
    # See output below, A..G, not A..H
    set r [string range $state(sha256c) 0 end-4]
    unset state
    return $r
}

# -------------------------------------------------------------------------
# HMAC Hashed Message Authentication (RFC 2104)
#
# hmac = H(K xor opad, H(K xor ipad, text))
#

# HMACInit --
#
#    This is equivalent to the SHA1Init procedure except that a key is
#    added into the algorithm
#
proc ::sha2::HMACInit {K} {

    # Key K is adjusted to be 64 bytes long. If K is larger, then use
    # the SHA1 digest of K and pad this instead.
    set len [string length $K]
    if {$len > 64} {
        set tok [SHA256Init]
        SHA256Update $tok $K
        set K [SHA256Final $tok]
        set len [string length $K]
    }
    set pad [expr {64 - $len}]
    append K [string repeat \0 $pad]

    # Cacluate the padding buffers.
    set Ki {}
    set Ko {}
    binary scan $K i16 Ks
    foreach k $Ks {
        append Ki [binary format i [expr {$k ^ 0x36363636}]]
        append Ko [binary format i [expr {$k ^ 0x5c5c5c5c}]]
    }

    set tok [SHA256Init]
    SHA256Update $tok $Ki;                 # initialize with the inner pad
    
    # preserve the Ko value for the final stage.
    # FRINK: nocheck
    set [subst $tok](Ko) $Ko

    return $tok
}

# HMACUpdate --
#
#    Identical to calling SHA256Update
#
proc ::sha2::HMACUpdate {token data} {
    SHA256Update $token $data
    return
}

# HMACFinal --
#
#    This is equivalent to the SHA256Final procedure. The hash context is
#    closed and the binary representation of the hash result is returned.
#
proc ::sha2::HMACFinal {token} {
    upvar #0 $token state

    set tok [SHA256Init];                 # init the outer hashing function
    SHA256Update $tok $state(Ko);         # prepare with the outer pad.
    SHA256Update $tok [SHA256Final $token]; # hash the inner result
    return [SHA256Final $tok]
}

# -------------------------------------------------------------------------
# Description:
#  This is the core SHA1 algorithm. It is a lot like the MD4 algorithm but
#  includes an extra round and a set of constant modifiers throughout.
#
set ::sha2::SHA256Transform_body {
    variable K
    upvar #0 $token state

    # FIPS 180-2: 6.2.2 SHA-256 Hash computation.
    binary scan $msg I* blocks
    set blockLen [llength $blocks]
    for {set i 0} {$i < $blockLen} {incr i 16} {
        set W [lrange $blocks $i [expr {$i+15}]]

        # FIPS 180-2: 6.2.2 (1) Prepare the message schedule
        # For t = 16 to 64 
        #   let Wt = (sigma1(Wt-2) + Wt-7 + sigma0(Wt-15) + Wt-16)
        set t2  13
        set t7   8
        set t15  0
        set t16 -1
        for {set t 16} {$t < 64} {incr t} {
            lappend W [expr {([sigma1 [lindex $W [incr t2]]] \
                                 + [lindex $W [incr t7]] \
                                 + [sigma0 [lindex $W [incr t15]]] \
                                 + [lindex $W [incr t16]]) & 0xffffffff}]
        }
        
        # FIPS 180-2: 6.2.2 (2) Initialise the working variables
        set A $state(A)
        set B $state(B)
        set C $state(C)
        set D $state(D)
        set E $state(E)
        set F $state(F)
        set G $state(G)
        set H $state(H)

        # FIPS 180-2: 6.2.2 (3) Do permutation rounds
        # For t = 0 to 63 do
        #   T1 = h + SIGMA1(e) + Ch(e,f,g) + Kt + Wt
        #   T2 = SIGMA0(a) + Maj(a,b,c)
        #   h = g; g = f;  f = e;  e = d + T1;  d = c;  c = b; b = a;
        #   a = T1 + T2
        #
        for {set t 0} {$t < 64} {incr t} {
            set T1 [expr {($H + [SIGMA1 $E] + [Ch $E $F $G] 
                          + [lindex $K $t] + [lindex $W $t]) & 0xffffffff}]
            set T2 [expr {([SIGMA0 $A] + [Maj $A $B $C]) & 0xffffffff}]
            set H $G
            set G $F
            set F $E
            set E [expr {($D + $T1) & 0xffffffff}]
            set D $C
            set C $B
            set B $A
            set A [expr {($T1 + $T2) & 0xffffffff}]
        }

        # FIPS 180-2: 6.2.2 (4) Compute the intermediate hash
        incr state(A) $A
        incr state(B) $B
        incr state(C) $C
        incr state(D) $D
        incr state(E) $E
        incr state(F) $F
        incr state(G) $G
        incr state(H) $H
    }

    return
}

# -------------------------------------------------------------------------

# FIPS 180-2: 4.1.2 equation 4.2
proc ::sha2::Ch {x y z} {
    return [expr {($x & $y) ^ (~$x & $z)}]
}

# FIPS 180-2: 4.1.2 equation 4.3
proc ::sha2::Maj {x y z} {
    return [expr {($x & $y) ^ ($x & $z) ^ ($y & $z)}]
}

# FIPS 180-2: 4.1.2 equation 4.4
#  (x >>> 2) ^ (x >>> 13) ^ (x >>> 22)
proc ::sha2::SIGMA0 {x} {
    return [expr {[>>> $x 2] ^ [>>> $x 13] ^ [>>> $x 22]}]
}

# FIPS 180-2: 4.1.2 equation 4.5
#  (x >>> 6) ^ (x >>> 11) ^ (x >>> 25)
proc ::sha2::SIGMA1 {x} {
    return [expr {[>>> $x 6] ^ [>>> $x 11] ^ [>>> $x 25]}]
}

# FIPS 180-2: 4.1.2 equation 4.6
#  s0 = (x >>> 7)  ^ (x >>> 18) ^ (x >> 3)
proc ::sha2::sigma0 {x} {
    #return [expr {[>>> $x 7] ^ [>>> $x 18] ^ (($x >> 3) & 0x1fffffff)}]
    return [expr {((($x<<25) | (($x>>7) & (0x7FFFFFFF>>6))) \
                 ^ (($x<<14) | (($x>>18) & (0x7FFFFFFF>>17))) & 0xFFFFFFFF) \
                 ^ (($x>>3) & 0x1fffffff)}]
}

# FIPS 180-2: 4.1.2 equation 4.7
#  s1 = (x >>> 17) ^ (x >>> 19) ^ (x >> 10)
proc ::sha2::sigma1 {x} {
    #return [expr {[>>> $x 17] ^ [>>> $x 19] ^ (($x >> 10) & 0x003fffff)}]
    return [expr {((($x<<15) | (($x>>17) & (0x7FFFFFFF>>16))) \
                 ^ (($x<<13) | (($x>>19) & (0x7FFFFFFF>>18))) & 0xFFFFFFFF) \
                 ^ (($x >> 10) & 0x003fffff)}]
}

# 32bit rotate-right
proc ::sha2::>>> {v n} {
    return [expr {(($v << (32 - $n)) \
                       | (($v >> $n) & (0x7FFFFFFF >> ($n - 1)))) \
                      & 0xFFFFFFFF}]
}

# 32bit rotate-left
proc ::sha2::<<< {v n} {
    return [expr {((($v << $n) \
                        | (($v >> (32 - $n)) \
                               & (0x7FFFFFFF >> (31 - $n))))) \
                      & 0xFFFFFFFF}]
}

# -------------------------------------------------------------------------
# We speed up the SHA256Transform code while maintaining readability in the
# source code by substituting inline for a number of functions.
# The idea is to reduce the number of [expr] calls.

# Inline the Ch function
regsub -all -line \
    {\[Ch (\$[ABCDEFGH]) (\$[ABCDEFGH]) (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((\1 \& \2) ^ ((~\1) \& \3))} \
    ::sha2::SHA256Transform_body

# Inline the Maj function
regsub -all -line \
    {\[Maj (\$[ABCDEFGH]) (\$[ABCDEFGH]) (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((\1 \& \2) ^ (\1 \& \3) ^ (\2 \& \3))} \
    ::sha2::SHA256Transform_body


# Inline the SIGMA0 function
regsub -all -line \
    {\[SIGMA0 (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((((\1<<30) | ((\1>>2) \& (0x7FFFFFFF>>1))) \& 0xFFFFFFFF) \
          ^ (((\1<<19) | ((\1>>13) \& (0x7FFFFFFF>>12))) \& 0xFFFFFFFF) \
          ^ (((\1<<10) | ((\1>>22) \& (0x7FFFFFFF>>21))) \& 0xFFFFFFFF) \
          )} \
    ::sha2::SHA256Transform_body

# Inline the SIGMA1 function
regsub -all -line \
    {\[SIGMA1 (\$[ABCDEFGH])\]} \
    $::sha2::SHA256Transform_body \
    {((((\1<<26) | ((\1>>6) \& (0x7FFFFFFF>>5))) \& 0xFFFFFFFF) \
          ^ (((\1<<21) | ((\1>>11) \& (0x7FFFFFFF>>10))) \& 0xFFFFFFFF) \
          ^ (((\1<<7) | ((\1>>25) \& (0x7FFFFFFF>>24))) \& 0xFFFFFFFF) \
          )} \
    ::sha2::SHA256Transform_body

proc ::sha2::SHA256Transform {token msg} $::sha2::SHA256Transform_body

# -------------------------------------------------------------------------

# Convert a integer value into a binary string in big-endian order.
proc ::sha2::byte {n v} {expr {((0xFF << (8 * $n)) & $v) >> (8 * $n)}}
proc ::sha2::bytes {v} { 
    #format %c%c%c%c [byte 3 $v] [byte 2 $v] [byte 1 $v] [byte 0 $v]
    format %c%c%c%c \
        [expr {((0xFF000000 & $v) >> 24) & 0xFF}] \
        [expr {(0xFF0000 & $v) >> 16}] \
        [expr {(0xFF00 & $v) >> 8}] \
        [expr {0xFF & $v}]
}

# -------------------------------------------------------------------------

proc ::sha2::Hex {data} {
    binary scan $data H* result
    return $result
}

# -------------------------------------------------------------------------

# Description:
#  Pop the nth element off a list. Used in options processing.
#
proc ::sha2::Pop {varname {nth 0}} {
    upvar $varname args
    set r [lindex $args $nth]
    set args [lreplace $args $nth $nth]
    return $r
}

# -------------------------------------------------------------------------

# fileevent handler for chunked file hashing.
#
proc ::sha2::Chunk {token channel {chunksize 4096}} {
    upvar #0 $token state
    
    SHA256Update $token [read $channel $chunksize]

    if {[eof $channel]} {
        fileevent $channel readable {}
        set state(reading) 0
    }
    return
}

# -------------------------------------------------------------------------

proc ::sha2::_sha256 {ver args} {
    array set opts {-hex 0 -filename {} -channel {} -chunksize 4096}
    if {[llength $args] == 1} {
        set opts(-hex) 1
    } else {
        while {[string match -* [set option [lindex $args 0]]]} {
            switch -glob -- $option {
                -hex       { set opts(-hex) 1 }
                -bin       { set opts(-hex) 0 }
                -file*     { set opts(-filename) [Pop args 1] }
                -channel   { set opts(-channel) [Pop args 1] }
                -chunksize { set opts(-chunksize) [Pop args 1] }
                default {
                    if {[llength $args] == 1} { break }
                    if {[string compare $option "--"] == 0} { Pop args; break }
                    set err [join [lsort [concat -bin [array names opts]]] ", "]
                    return -code error "bad option $option:\
                    must be one of $err"
                }
            }
            Pop args
        }
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {
        if {[llength $args] != 1} {
            return -code error "wrong # args: should be\
                \"[namespace current]::sha$ver ?-hex|-bin? -filename file\
                | -channel channel | string\""
        }
        set tok [SHA${ver}Init]
        SHA${ver}Update $tok [lindex $args 0]
        set r [SHA${ver}Final $tok]

    } else {

        set tok [SHA${ver}Init]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        # FRINK: nocheck
        vwait [subst $tok](reading)
        set r [SHA${ver}Final $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

interp alias {} ::sha2::sha256 {} ::sha2::_sha256 256
interp alias {} ::sha2::sha224 {} ::sha2::_sha256 224

# -------------------------------------------------------------------------

proc ::sha2::hmac {args} {
    array set opts {-hex 1 -filename {} -channel {} -chunksize 4096}
    if {[llength $args] != 2} {
        while {[string match -* [set option [lindex $args 0]]]} {
            switch -glob -- $option {
                -key       { set opts(-key) [Pop args 1] }
                -hex       { set opts(-hex) 1 }
                -bin       { set opts(-hex) 0 }
                -file*     { set opts(-filename) [Pop args 1] }
                -channel   { set opts(-channel) [Pop args 1] }
                -chunksize { set opts(-chunksize) [Pop args 1] }
                default {
                    if {[llength $args] == 1} { break }
                    if {[string compare $option "--"] == 0} { Pop args; break }
                    set err [join [lsort [array names opts]] ", "]
                    return -code error "bad option $option:\
                    must be one of $err"
                }
            }
            Pop args
        }
    }

    if {[llength $args] == 2} {
        set opts(-key) [Pop args]
    }

    if {![info exists opts(-key)]} {
        return -code error "wrong # args:\
            should be \"hmac ?-hex? -key key -filename file | string\""
    }

    if {$opts(-filename) != {}} {
        set opts(-channel) [open $opts(-filename) r]
        fconfigure $opts(-channel) -translation binary
    }

    if {$opts(-channel) == {}} {

        if {[llength $args] != 1} {
            return -code error "wrong # args:\
                should be \"hmac ?-hex? -key key -filename file | string\""
        }
        set tok [HMACInit $opts(-key)]
        HMACUpdate $tok [lindex $args 0]
        set r [HMACFinal $tok]

    } else {

        set tok [HMACInit $opts(-key)]
        # FRINK: nocheck
        set [subst $tok](reading) 1
        fileevent $opts(-channel) readable \
            [list [namespace origin Chunk] \
                 $tok $opts(-channel) $opts(-chunksize)]
        # FRINK: nocheck
        vwait [subst $tok](reading)
        set r [HMACFinal $tok]

        # If we opened the channel - we should close it too.
        if {$opts(-filename) != {}} {
            close $opts(-channel)
        }
    }
    
    if {$opts(-hex)} {
        set r [Hex $r]
    }
    return $r
}

# -------------------------------------------------------------------------

# Try and load a compiled extension to help.
namespace eval ::sha2 {
    variable e {}
    foreach e [KnownImplementations] {
	if {[LoadAccelerator $e]} {
	    SwitchTo $e
	    break
	}
    }
    unset e
}

package provide sha256 1.0.4

# -------------------------------------------------------------------------
# Local Variables:
#   mode: tcl
#   indent-tabs-mode: nil
# End:

# THATCH /pkgs/tcllib/base64/pkgIndex.tcl size 360 mtime 1659183355
if {![package vsatisfies [package provide Tcl] 8.2]} {return}
package ifneeded base64   2.5 [list source [file join $dir base64.tcl]]
package ifneeded uuencode 1.1.5 [list source [file join $dir uuencode.tcl]]
package ifneeded yencode  1.1.3 [list source [file join $dir yencode.tcl]]
package ifneeded ascii85  1.0   [list source [file join $dir ascii85.tcl]]

# THATCH /pkgs/tcllib/base64/base64.tcl size 12080 mtime 1659183355
# base64.tcl --
#
# Encode/Decode base64 for a string
# Stephen Uhler / Brent Welch (c) 1997 Sun Microsystems
# The decoder was done for exmh by Chris Garrigues
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Version 1.0   implemented Base64_Encode, Base64_Decode
# Version 2.0   uses the base64 namespace
# Version 2.1   fixes various decode bugs and adds options to encode
# Version 2.2   is much faster, Tcl8.0 compatible
# Version 2.2.1 bugfixes
# Version 2.2.2 bugfixes
# Version 2.3   bugfixes and extended to support Trf
# Version 2.4.x bugfixes

# @mdgen EXCLUDE: base64c.tcl

package require Tcl 8.2
namespace eval ::base64 {
    namespace export encode decode
}

package provide base64 2.5

if {[package vsatisfies [package require Tcl] 8.6]} {
    proc ::base64::encode {args} {
	binary encode base64 -maxlen 76 {*}$args
    }

    proc ::base64::decode {string} {
	# Tcllib is strict with respect to end of input, yet lax for
	# invalid characters outside of that.
	regsub -all -- {[^ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/]} $string {} string
	binary decode base64 -strict $string
    }

    return
}

if {![catch {package require Trf 2.0}]} {
    # Trf is available, so implement the functionality provided here
    # in terms of calls to Trf for speed.

    # ::base64::encode --
    #
    #	Base64 encode a given string.
    #
    # Arguments:
    #	args	?-maxlen maxlen? ?-wrapchar wrapchar? string
    #
    #		If maxlen is 0, the output is not wrapped.
    #
    # Results:
    #	A Base64 encoded version of $string, wrapped at $maxlen characters
    #	by $wrapchar.

    proc ::base64::encode {args} {
	# Set the default wrapchar and maximum line length to match
	# the settings for MIME encoding (RFC 3548, RFC 2045). These
	# are the settings used by Trf as well. Various RFCs allow for
	# different wrapping characters and wraplengths, so these may
	# be overridden by command line options.
	set wrapchar "\n"
	set maxlen 76

	if { [llength $args] == 0 } {
	    error "wrong # args: should be \"[lindex [info level 0] 0]\
		    ?-maxlen maxlen? ?-wrapchar wrapchar? string\""
	}

	set optionStrings [list "-maxlen" "-wrapchar"]
	for {set i 0} {$i < [llength $args] - 1} {incr i} {
	    set arg [lindex $args $i]
	    set index [lsearch -glob $optionStrings "${arg}*"]
	    if { $index == -1 } {
		error "unknown option \"$arg\": must be -maxlen or -wrapchar"
	    }
	    incr i
	    if { $i >= [llength $args] - 1 } {
		error "value for \"$arg\" missing"
	    }
	    set val [lindex $args $i]

	    # The name of the variable to assign the value to is extracted
	    # from the list of known options, all of which have an
	    # associated variable of the same name as the option without
	    # a leading "-". The [string range] command is used to strip
	    # of the leading "-" from the name of the option.
	    #
	    # FRINK: nocheck
	    set [string range [lindex $optionStrings $index] 1 end] $val
	}

	# [string is] requires Tcl8.2; this works with 8.0 too
	if {[catch {expr {$maxlen % 2}}]} {
	    return -code error "expected integer but got \"$maxlen\""
	} elseif {$maxlen < 0} {
	    return -code error "expected positive integer but got \"$maxlen\""
	}

	set string [lindex $args end]
	set result [::base64 -mode encode -- $string]

	# Trf's encoder implicitly uses the settings -maxlen 76,
	# -wrapchar \n for its output. We may have to reflow this for
	# the settings chosen by the user. A second difference is that
	# Trf closes the output with the wrap char sequence,
	# always. The code here doesn't. Therefore 'trimright' is
	# needed in the fast cases.

	if {($maxlen == 76) && [string equal $wrapchar \n]} {
	    # Both maxlen and wrapchar are identical to Trf's
	    # settings. This is the super-fast case, because nearly
	    # nothing has to be done. Only thing to do is strip a
	    # terminating wrapchar.
	    set result [string trimright $result]
	} elseif {$maxlen == 76} {
	    # wrapchar has to be different here, length is the
	    # same. We can use 'string map' to transform the wrap
	    # information.
	    set result [string map [list \n $wrapchar] \
			    [string trimright $result]]
	} elseif {$maxlen == 0} {
	    # Have to reflow the output to no wrapping. Another fast
	    # case using only 'string map'. 'trimright' is not needed
	    # here.

	    set result [string map [list \n ""] $result]
	} else {
	    # Have to reflow the output from 76 to the chosen maxlen,
	    # and possibly change the wrap sequence as well.

	    # Note: After getting rid of the old wrap sequence we
	    # extract the relevant segments from the string without
	    # modifying the string. Modification, i.e. removal of the
	    # processed part, means 'shifting down characters in
	    # memory', making the algorithm O(n^2). By avoiding the
	    # modification we stay in O(n).

	    set result [string map [list \n ""] $result]
	    set l [expr {[string length $result]-$maxlen}]
	    for {set off 0} {$off < $l} {incr off $maxlen} {
		append res [string range $result $off [expr {$off+$maxlen-1}]] $wrapchar
	    }
	    append res [string range $result $off end]
	    set result $res
	}

	return $result
    }

    # ::base64::decode --
    #
    #	Base64 decode a given string.
    #
    # Arguments:
    #	string	The string to decode.  Characters not in the base64
    #		alphabet are ignored (e.g., newlines)
    #
    # Results:
    #	The decoded value.

    proc ::base64::decode {string} {
	regsub -all {\s} $string {} string
	::base64 -mode decode -- $string
    }

} else {
    # Without Trf use a pure tcl implementation

    namespace eval base64 {
	variable base64 {}
	variable base64_en {}

	# We create the auxiliary array base64_tmp, it will be unset later.
	variable base64_tmp
	variable i

	set i 0
	foreach char {A B C D E F G H I J K L M N O P Q R S T U V W X Y Z \
		a b c d e f g h i j k l m n o p q r s t u v w x y z \
		0 1 2 3 4 5 6 7 8 9 + /} {
	    set base64_tmp($char) $i
	    lappend base64_en $char
	    incr i
	}

	#
	# Create base64 as list: to code for instance C<->3, specify
	# that [lindex $base64 67] be 3 (C is 67 in ascii); non-coded
	# ascii chars get a {}. we later use the fact that lindex on a
	# non-existing index returns {}, and that [expr {} < 0] is true
	#

	# the last ascii char is 'z'
	variable char
	variable len
	variable val

	scan z %c len
	for {set i 0} {$i <= $len} {incr i} {
	    set char [format %c $i]
	    set val {}
	    if {[info exists base64_tmp($char)]} {
		set val $base64_tmp($char)
	    } else {
		set val {}
	    }
	    lappend base64 $val
	}

	# code the character "=" as -1; used to signal end of message
	scan = %c i
	set base64 [lreplace $base64 $i $i -1]

	# remove unneeded variables
	unset base64_tmp i char len val

	namespace export encode decode
    }

    # ::base64::encode --
    #
    #	Base64 encode a given string.
    #
    # Arguments:
    #	args	?-maxlen maxlen? ?-wrapchar wrapchar? string
    #
    #		If maxlen is 0, the output is not wrapped.
    #
    # Results:
    #	A Base64 encoded version of $string, wrapped at $maxlen characters
    #	by $wrapchar.

    proc ::base64::encode {args} {
	set base64_en $::base64::base64_en

	# Set the default wrapchar and maximum line length to match
	# the settings for MIME encoding (RFC 3548, RFC 2045). These
	# are the settings used by Trf as well. Various RFCs allow for
	# different wrapping characters and wraplengths, so these may
	# be overridden by command line options.
	set wrapchar "\n"
	set maxlen 76

	if { [llength $args] == 0 } {
	    error "wrong # args: should be \"[lindex [info level 0] 0]\
		    ?-maxlen maxlen? ?-wrapchar wrapchar? string\""
	}

	set optionStrings [list "-maxlen" "-wrapchar"]
	for {set i 0} {$i < [llength $args] - 1} {incr i} {
	    set arg [lindex $args $i]
	    set index [lsearch -glob $optionStrings "${arg}*"]
	    if { $index == -1 } {
		error "unknown option \"$arg\": must be -maxlen or -wrapchar"
	    }
	    incr i
	    if { $i >= [llength $args] - 1 } {
		error "value for \"$arg\" missing"
	    }
	    set val [lindex $args $i]

	    # The name of the variable to assign the value to is extracted
	    # from the list of known options, all of which have an
	    # associated variable of the same name as the option without
	    # a leading "-". The [string range] command is used to strip
	    # of the leading "-" from the name of the option.
	    #
	    # FRINK: nocheck
	    set [string range [lindex $optionStrings $index] 1 end] $val
	}

	# [string is] requires Tcl8.2; this works with 8.0 too
	if {[catch {expr {$maxlen % 2}}]} {
	    return -code error "expected integer but got \"$maxlen\""
	} elseif {$maxlen < 0} {
	    return -code error "expected positive integer but got \"$maxlen\""
	}

	set string [lindex $args end]

	set result {}
	set state 0
	set length 0


	# Process the input bytes 3-by-3

	binary scan $string c* X

	foreach {x y z} $X {
	    ADD [lindex $base64_en [expr {($x >>2) & 0x3F}]]
	    if {$y != {}} {
		ADD [lindex $base64_en [expr {(($x << 4) & 0x30) | (($y >> 4) & 0xF)}]]
		if {$z != {}} {
		    ADD [lindex $base64_en [expr {(($y << 2) & 0x3C) | (($z >> 6) & 0x3)}]]
		    ADD [lindex $base64_en [expr {($z & 0x3F)}]]
		} else {
		    set state 2
		    break
		}
	    } else {
		set state 1
		break
	    }
	}
	if {$state == 1} {
	    ADD [lindex $base64_en [expr {(($x << 4) & 0x30)}]]
	    ADD =
	    ADD =
	} elseif {$state == 2} {
	    ADD [lindex $base64_en [expr {(($y << 2) & 0x3C)}]]
	    ADD =
	}
	return $result
    }

    proc ::base64::ADD {x} {
	# The line length check is always done before appending so
	# that we don't get an extra newline if the output is a
	# multiple of $maxlen chars long.

	upvar 1 maxlen maxlen length length result result wrapchar wrapchar
	if {$maxlen && $length >= $maxlen} {
	    append result $wrapchar
	    set length 0
	}
	append result $x
	incr length
	return
    }

    # ::base64::decode --
    #
    #	Base64 decode a given string.
    #
    # Arguments:
    #	string	The string to decode.  Characters not in the base64
    #		alphabet are ignored (e.g., newlines)
    #
    # Results:
    #	The decoded value.

    proc ::base64::decode {string} {
	if {[string length $string] == 0} {return ""}

	set base64 $::base64::base64
	set output "" ; # Fix for [Bug 821126]
	set nums {}

	binary scan $string c* X
	lappend X 61 ;# force a terminator
	foreach x $X {
	    set bits [lindex $base64 $x]
	    if {$bits >= 0} {
		if {[llength [lappend nums $bits]] == 4} {
		    foreach {v w z y} $nums break
		    set a [expr {($v << 2) | ($w >> 4)}]
		    set b [expr {(($w & 0xF) << 4) | ($z >> 2)}]
		    set c [expr {(($z & 0x3) << 6) | $y}]
		    append output [binary format ccc $a $b $c]
		    set nums {}
		}
	    } elseif {$bits == -1} {
		# = indicates end of data.  Output whatever chars are
		# left, if any.
		if {![llength $nums]} break
		# The encoding algorithm dictates that we can only
		# have 1 or 2 padding characters.  If x=={}, we must
		# (*) have 12 bits of input (enough for 1 8-bit
		# output).  If x!={}, we have 18 bits of input (enough
		# for 2 8-bit outputs).
		#
		# (*) If we don't then the input is broken (bug 2976290).

		foreach {v w z} $nums break

		# Bug 2976290
		if {$w == {}} {
		    return -code error "Not enough data to process padding"
		}

		set a [expr {($v << 2) | (($w & 0x30) >> 4)}]
		if {$z == {}} {
		    append output [binary format c $a ]
		} else {
		    set b [expr {(($w & 0xF) << 4) | (($z & 0x3C) >> 2)}]
		    append output [binary format cc $a $b]
		}
		break
	    } else {
		# RFC 2045 says that line breaks and other characters not part
		# of the Base64 alphabet must be ignored, and that the decoder
		# can optionally emit a warning or reject the message.  We opt
		# not to do so, but to just ignore the character.
		continue
	    }
	}
	return $output
    }
}

# # ## ### ##### ######## ############# #####################
return


# THATCH /pkgs/tcllib/license.terms size 2155 mtime 1659183355
This software is copyrighted by Ajuba Solutions and other parties.
The following terms apply to all files associated with the software
unless explicitly disclaimed in individual files.

The authors hereby grant permission to use, copy, modify, distribute,
and license this software and its documentation for any purpose,
provided that existing copyright notices are retained in all copies
and that this notice is included verbatim in any distributions. No
written agreement, license, or royalty fee is required for any of the
authorized uses.  Modifications to this software may be copyrighted by
their authors and need not follow the licensing terms described here,
provided that the new terms are clearly indicated on the first page of
each file where they apply.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
NON-INFRINGEMENT.  THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND
THE AUTHORS AND DISTRIBUTORS HAVE NO OBLIGATION TO PROVIDE
MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

GOVERNMENT USE: If you are acquiring this software on behalf of the
U.S. government, the Government shall have only "Restricted Rights" in
the software and related documentation as defined in the Federal
Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
are acquiring the software on behalf of the Department of Defense, the
software shall be classified as "Commercial Computer Software" and the
Government shall have only "Restricted Rights" as defined in Clause
252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
authors grant the U.S. Government and others acting in its behalf
permission to use and distribute the software in accordance with the
terms specified in this license.
